from copy import deepcopy

from collections import defaultdict
from typing import List, Tuple, Callable, Union
import logging

import numpy as np
import pandas as pd
import numpy.typing as npt
from scipy.special import softmax
from sklearn.utils import resample

import torch
from torch import nn
from torch.utils.data import DataLoader
from torch.optim.swa_utils import AveragedModel

from nrn.modules import BanditNRNModule
from .pruningrn import PruningReasoningNetworkModel
from nrn.utils.operations import register_hooks


class BaseBanditNRNModel(PruningReasoningNetworkModel):

    def __init__(
            self,
            target_names: List[str],
            feature_names: List[str],
            input_size: int,
            output_size: int,
            layer_sizes: List[int],
            n_selected_features_input: int,
            n_selected_features_internal: int,
            n_selected_features_output: int,
            perform_prune_quantile: float,
            ucb_scale: float,
            normal_form: str = 'dnf',
            delta: float = 2.0,
            prune_strategy: str = 'class',
            bootstrap: bool = True,
            swa: bool = False,
            add_negations: bool = False,
            weight_init: float = 0.2,
            policy_init: torch.Tensor = None,
            logits: bool = False
    ):
        """
        Initialize a Bandit Reinforced Reasoning Network model.

        Example:
            model = BaseBanditRRNModel(
                target_names=['class1', 'class2'],
                feature_names=['feature1', 'feature2', 'feature3'],
                input_size=3,
                output_size=2,
                layer_sizes=[3, 3]
                n_selected_features_input=2,
                n_selected_features_internal=2,
                n_selected_features_output=1,
                ucb_scale=1.96,
                perform_prune_quantile=0.7,
                normal_form='dnf',
                prune_strategy='class',
                bootstrap=False,
                swa=False
            )

        Args:
            target_names (list): A list of the target names.
            feature_names (list): A list of feature names.
            input_size (int): number of features from input.
            output_size (int): number of outputs.
            layer_sizes (list): A list containing the number of output logics for each layer.
            n_selected_features_input (int): The number of features to include in each logic in the input layer.
            n_selected_features_internal (int): The number of logics to include in each logic in the internal layers.
            n_selected_features_output (int): The number of logics to include in each logic in the output layer.
            perform_prune_quantile (float): The quantile to use for pruning randomized rn.
            ucb_scale (float): The scale of the confidence interval in the multi-armed bandit policy.
                               c = 1.96 is a 95% confidence interval.
            normal_form (str): 'dnf' for disjunctive normal form network; 'cnf' for conjunctive normal form network.
            delta (float): higher values increase diversity of logic generation away from existing logics.
            prune_strategy (str): Either 'class' or 'logic'.  Determines which pruning strategy to use.
            bootstrap (bool): Use boostrap samples to evaluate each atomic logic in logic prune strategy.
            swa (bool): Use stochastic weight averaging
            add_negations (bool): add negations of logic.
            weight_init (float): Upper bound of uniform weight initialization.  Lower bound is negated value.
        """
        PruningReasoningNetworkModel.__init__(self)
        assert n_selected_features_input <= len(feature_names), \
            "`n_selected_features_input` must be <= number of features"
        assert n_selected_features_internal <= min(layer_sizes), \
            "`n_selected_features_internal` must be <= min(layer_sizes)"
        assert n_selected_features_output <= layer_sizes[-1], \
            "`n_selected_features_output` must be <= layer_sizes[-1]"
        if not add_negations:
            assert n_selected_features_input > 1, "`n_selected_features_input` must be greater than 1."
        assert n_selected_features_internal > 1, "`n_selected_features_internal` must be greater than 1."
        assert n_selected_features_output > 1, "`n_selected_features_output` must be greater than 1."

        self.target_names = target_names
        self.feature_names = feature_names
        self.input_size = input_size
        self.output_size = output_size
        self.n_selected_features_input = n_selected_features_input
        self.n_selected_features_internal = n_selected_features_internal
        self.n_selected_features_output = n_selected_features_output
        self.perform_prune_quantile = perform_prune_quantile
        self.ucb_scale = ucb_scale
        self.normal_form = normal_form
        self.delta = delta
        self.prune_strategy = prune_strategy
        self.bootstrap = bootstrap
        self.swa = swa
        self.add_negations = add_negations
        self.weight_init = weight_init
        self.logits = logits

        self.rn = BanditNRNModule(
            input_size=input_size,
            output_size=output_size,
            layer_sizes=layer_sizes,
            feature_names=feature_names,
            n_selected_features_input=n_selected_features_input,
            n_selected_features_internal=n_selected_features_internal,
            n_selected_features_output=n_selected_features_output,
            perform_prune_quantile=perform_prune_quantile,
            ucb_scale=ucb_scale,
            normal_form=normal_form,
            add_negations=add_negations,
            weight_init=weight_init,
            logits=logits
        )

        self.averaged_rn = None
        if self.swa:
            self.averaged_rn = AveragedModel(self.rn)

        self.target_names = target_names
        self.bootstrap_sample_map = {k: defaultdict(list) for k in range(self.output_size)}

        if torch.cuda.device_count() > 1:
            self.logger.info(f"Using {torch.cuda.device_count()} GPUs!")
            self.rn = nn.DataParallel(self.rn)
        if self.USE_CUDA:
            self.logger.info(f"Using GPU")
            self.rn = self.rn.cuda()
        elif self.USE_MPS:
            self.logger.info(f"Using MPS")
            self.rn = self.rn.to('mps')

        self.USE_DATA_PARALLEL = isinstance(self.rn, torch.nn.DataParallel)

        # initialize uniform policy
        self.predicates_policy = torch.empty((output_size, input_size)).uniform_(0, 1)
        if policy_init is not None:
            self.predicates_policy.copy_(policy_init)
        # self.register_buffer('predicates_policy', predicates_policy)
        self.predicates_rewards_history = self._initialize_rewards_history()

        self.logger = logging.getLogger(self.__class__.__name__)

    def _initialize_rewards_history(self) -> dict:
        """
        Initialize the reward histories for the multi-armed bandit.

        Returns:
            dict: dictionary of pd.DataFrame with rewards histories by class.
        """
        predicates_rewards_history = {}
        for output in range(self.output_size):
            predicates_rewards_history[output] = pd.DataFrame(
                {'predicates': range(self.input_size), 'rewards': self.predicates_policy[output].cpu().numpy()})
        return predicates_rewards_history

    def _get_activations(self, dl: DataLoader) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """"
        Get logic activations, targets, predictions and sample indexes used for 'logic' pruning strategy.

        Args:
            dl (DataLoader): PyTorch data loader with training data.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
                logic activations, targets, sample indexes
        """
        all_sample_idx = []
        all_activations = []
        all_targets = []
        activations = {}
        register_hooks(self, activations)
        for batch in dl:
            # [BATCH_SIZE, N_FEATURES]
            features = batch['features']
            # [BATCH_SIZE, N_TARGETS]
            target = batch['target']
            sample_idx = batch['sample_idx']

            if target.ndim > 2:
                target = target.squeeze(-1)

            if self.USE_CUDA:
                features = features.cuda()
                target = target.cuda()
            elif self.USE_MPS:
                features = features.to('mps')
                target = target.to('mps')

            # [BATCH_SIZE, N_TARGETS]
            if self.USE_DATA_PARALLEL:
                _ = self.rn.module.model[0](features)
            else:
                _ = self.rn.model[0](features)

            all_activations += [activations['0']]
            all_targets += [deepcopy(target)]
            all_sample_idx += [deepcopy(sample_idx)]

        all_activations = torch.cat(all_activations, dim=0)
        all_targets = torch.cat(all_targets, dim=0)
        all_sample_idx = torch.cat(all_sample_idx, dim=0)

        return all_activations, all_targets, all_sample_idx

    def _evaluate_activations(
            self,
            targets: torch.Tensor,
            activations: torch.Tensor,
            sample_idx: torch.Tensor,
            input_layer_weights: torch.Tensor,
            internal_layer_mask: torch.Tensor,
            mask: torch.Tensor,
            class_indices: torch.Tensor,
            output_metric: Callable,
            rewards: torch.Tensor = None,
            dimension: int = -2,
            mode: str = 'identify_important_features',
            objective: str = 'maximize'
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Evaluate activations of logic to determine important features or logic to prune.

        Args:
            targets (torch.Tensor): targets for training data.
            activations (torch.Tensor): activations for input layer logic from forward pass on training data.
            sample_idx (torch.Tensor): sample indexes from forward pass on training data.
            input_layer_weights (torch.Tensor): Detached input layer weights from the model.
            internal_layer_mask (torch.Tensor): Mask of which input layer logics are used by the network.
            mask (torch.Tensor): mask to be modified by process.
            class_indices (list): List of integers indicating which class policies to update.
            output_metric (sklearn.metrics): Scikit-learn performance metric to use for evaluating logic.
            rewards (torch.Tensor): rewards if any.
            dimension (int): Dimension corresponding to logics from input_layer.
            mode (str): 'identify_important_features' or 'prune'.
            objective (str): 'maximize' or 'minimize'.  Which direction is the optimal performance for output_metric.

         Returns:
            Tuple[torch.Tensor, torch.Tensor]: boolean mask of important features or features to prune, rewards
        """
        sample_idx = sample_idx.cpu().numpy()
        targets = targets.cpu().numpy()
        activations = activations.cpu().numpy()

        train_size = targets.shape[0]
        indices = list(range(train_size))
        sample_idx_to_data_idx = {k: v for k, v in zip(sample_idx, indices)}

        for i in range(self.output_size):
            if not (i in class_indices):
                mask[i] = False
            else:

                performances = []
                for j in range(input_layer_weights.size(dimension)):

                    if torch.isin(torch.tensor(j), internal_layer_mask[i].cpu()):

                        if self.bootstrap:
                            # ensures that the boostrap samples are the same for each logic, every time
                            if j not in self.bootstrap_sample_map[i]:
                                bootstrap_sample_idx = resample(sample_idx, stratify=targets).tolist()
                                self.bootstrap_sample_map[i][j] = bootstrap_sample_idx
                            train_prune_idx = [sample_idx_to_data_idx[x] for x in self.bootstrap_sample_map[i][j]]
                        else:
                            train_prune_idx = indices

                        pos_performance = output_metric(
                            targets[train_prune_idx, i].reshape(-1, 1),
                            activations[train_prune_idx, i, 0, j].reshape(-1, 1))
                        neg_performance = output_metric(
                            targets[train_prune_idx, i].reshape(-1, 1),
                            activations[train_prune_idx, i, 0, j].reshape(-1, 1) * -1)

                        if pos_performance < 0 or neg_performance < 0:
                            raise AssertionError(
                                "Performance metric produced value less than zero. Must produce a positive value.")

                        if objective == 'maximize':
                            performances += [max(pos_performance, neg_performance)]
                        elif objective == 'minimize':
                            performances += [min(pos_performance, neg_performance)]
                        else:
                            raise ValueError("'objective' must be 'maximize' or 'minimize'.")
                    else:
                        if objective == 'maximize':
                            performances += [0]
                        elif objective == 'minimize':
                            performances += [100000]

                performances = np.array(performances)
                if objective == 'minimize' and any(performances > 0):
                    performances = np.log1p(np.clip(1 / performances + 1e-8, 0.0, 100000))

                performances_quantile_threshold = np.quantile(
                    performances, self.perform_prune_quantile)
                if len(performances) > 0:
                    for j in range(input_layer_weights.size(dimension)):
                        if torch.isin(torch.tensor(j), internal_layer_mask[i].cpu()):
                            if mode == 'identify_important_features':
                                if performances[j] >= performances_quantile_threshold:
                                    mask[i, j, :] = True
                                    if rewards is not None:
                                        rewards[i, j, :] = torch.tensor(performances[j]).repeat(rewards[i, j, :].size())
                            elif mode == 'prune':
                                if performances[j] < performances_quantile_threshold:
                                    mask[i, :, j] = True
                            else:
                                raise ValueError("'mode' must be 'identify_important_features' or 'prune'.")

        if mode == 'prune':
            mask = mask.bool().transpose(-2, -1)

        return mask, rewards

    def _evaluate_weights(
            self,
            input_layer_weights: torch.Tensor,
            class_indices: Union[List[int], torch.Tensor],
            mode: str = 'identify_important_features'
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Perform weight evaluation to determine which logic to prune or which features are important.

        Args:
            input_layer_weights (torch.Tensor): Detached input layer weights from the model.
            class_indices (list): List of integers indicating which class policies to update.
            mode (str): 'identify_important_features' or 'prune'.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: boolean mask of important features or features to prune, rewards
        """
        input_layer_weights_abs = torch.abs(input_layer_weights)

        if mode == 'identify_important_features':
            try:
                quantile_mask = torch.tensor(
                    input_layer_weights_abs >= input_layer_weights_abs.data.reshape(
                        self.output_size, -1).quantile(
                        self.perform_prune_quantile, dim=1,
                        keepdim=False).unsqueeze(-1).unsqueeze(-1).expand(input_layer_weights.size()))
            except NotImplementedError as e:
                quantile_mask = torch.Tensor(
                    (input_layer_weights_abs.to('cpu') >= input_layer_weights_abs.to('cpu').reshape(
                        self.output_size, -1).quantile(
                        self.perform_prune_quantile, dim=1,
                        keepdim=False).unsqueeze(-1).unsqueeze(-1).expand(input_layer_weights.size())).clone().detach())

        elif mode == 'prune':

            try:
                quantile_mask = torch.tensor(
                    input_layer_weights_abs <= input_layer_weights_abs.data.reshape(self.output_size, -1).quantile(
                        self.perform_prune_quantile, dim=1,
                        keepdim=False).unsqueeze(-1).unsqueeze(-1).expand(input_layer_weights_abs.size()))
            except NotImplementedError as e:
                quantile_mask = torch.Tensor(
                    (input_layer_weights_abs.to('cpu') <= input_layer_weights_abs.to('cpu').reshape(
                        self.output_size, -1).quantile(
                        self.perform_prune_quantile, dim=1,
                        keepdim=False).unsqueeze(-1).unsqueeze(-1).expand(
                        input_layer_weights_abs.size())).clone().detach())
        else:
            raise ValueError("'mode' must be 'identify_important_features' or 'prune'.")

        for i in range(self.output_size):
            if not (i in class_indices):
                quantile_mask[i] = False

        if mode == 'identify_important_features':
            prune_mask = quantile_mask
            rewards = input_layer_weights_abs
        elif mode == 'prune':
            prune_mask = quantile_mask.bool().transpose(-2, -1)
            rewards = None

        return prune_mask, rewards

    def _evaluate_logic_weights(
            self,
            input_layer_weights: torch.Tensor,
            internal_layer_weights: torch.Tensor,
            internal_layer_mask: torch.Tensor,
            class_indices: Union[List[int], torch.Tensor],
            mode: str = 'identify_important_features'
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Perform weight evaluation to determine which logic to prune or which features are important.

        Args:
            input_layer_weights (torch.Tensor): Detached input layer weights from the model.
            class_indices (list): List of integers indicating which class policies to update.
            mode (str): 'identify_important_features' or 'prune'.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: boolean mask of important features or features to prune, rewards
        """
        logic_layer_weights_abs = torch.abs(internal_layer_weights)

        if mode == 'identify_important_features':
            try:
                quantile_mask = torch.tensor(
                    logic_layer_weights_abs >= logic_layer_weights_abs.data.reshape(
                        self.output_size, -1).quantile(
                        self.perform_prune_quantile, dim=1,
                        keepdim=False).unsqueeze(-1).unsqueeze(-1).expand(internal_layer_weights.size()))
            except NotImplementedError as e:
                quantile_mask = torch.Tensor(
                    (logic_layer_weights_abs.to('cpu') >= logic_layer_weights_abs.to('cpu').reshape(
                        self.output_size, -1).quantile(
                        self.perform_prune_quantile, dim=1,
                        keepdim=False).unsqueeze(-1).unsqueeze(-1).expand(internal_layer_weights.size())).clone().detach())

        elif mode == 'prune':

            try:
                quantile_mask = torch.tensor(
                    logic_layer_weights_abs <= logic_layer_weights_abs.data.reshape(self.output_size, -1).quantile(
                        self.perform_prune_quantile, dim=1,
                        keepdim=False).unsqueeze(-1).unsqueeze(-1).expand(logic_layer_weights_abs.size()))
            except NotImplementedError as e:
                quantile_mask = torch.Tensor(
                    (logic_layer_weights_abs.to('cpu') <= logic_layer_weights_abs.to('cpu').reshape(
                        self.output_size, -1).quantile(
                        self.perform_prune_quantile, dim=1,
                        keepdim=False).unsqueeze(-1).unsqueeze(-1).expand(
                        logic_layer_weights_abs.size())).clone().detach())
        else:
            raise ValueError("'mode' must be 'identify_important_features' or 'prune'.")

        for i in range(self.output_size):
            if not (i in class_indices):
                quantile_mask[i] = False

        # is the logic used anywhere?
        # iterate over classes.  take the max weight and max quantile mask from anywhere it's used
        prune_mask = torch.zeros_like(input_layer_weights).bool()
        rewards = torch.zeros_like(input_layer_weights)
        for c in range(self.output_size):
            for i in range(internal_layer_mask.size(1)):
                for k, j in enumerate(internal_layer_mask[c, i, :]):
                    if prune_mask[c, :, j].max() < quantile_mask[c, i, k]:
                        prune_mask[c, :, j] = quantile_mask[c, i, k]
                    if (0 < rewards[c, :, j].max() < internal_layer_weights[c, i, k]) or rewards[c, :, j].max() == 0:
                        rewards[c, :, j] = internal_layer_weights[c, i, k]

        if mode == 'identify_important_features':
            prune_mask = prune_mask.transpose(-2, -1)
            rewards = rewards.transpose(-2, -1)
        elif mode == 'prune':
            prune_mask = prune_mask.transpose(-2, -1)
            rewards = None

        return prune_mask, rewards

    def _identify_important_features(
            self,
            dl: DataLoader = None,
            class_indices: List[int] = None,
            output_metric: Callable = None,
            objective: str = 'maximize'
    ) -> Tuple[List[List[int]], List[List[float]]]:
        """
        Identify important features.

        Args:
            dl (DataLoader): PyTorch data loader with training data.
            class_indices (list): List of integers indicating which class policies to update.
            output_metric (sklearn.metrics): Scikit-learn performance metric to use for evaluating logic.
            objective (str): 'maximize' or 'minimize'.  Which direction is the optimal performance for output_metric.

        Returns:
            Tuple[List[List[int]], List[List[float]]]: important feature indexes, important feature rewards
        """
        if self.USE_DATA_PARALLEL:
            input_layer_mask = self.rn.module.model[0].mask
            input_layer_weights = torch.abs(self.rn.module.model[0].weights.clone().detach().transpose(-2, -1))
            if len(self.rn.module.model) > 1:
                internal_layer_mask = self.rn.module.model[1].mask
                internal_layer_weights = torch.abs(self.rn.module.model[1].weights.clone().detach().transpose(-2, -1))
            else:
                internal_layer_mask = self.rn.module.output_layer.mask
                internal_layer_weights = torch.abs(self.rn.module.output_layer.weights.clone().detach().transpose(-2, -1))
        else:
            input_layer_mask = self.rn.model[0].mask
            input_layer_weights = torch.abs(self.rn.model[0].weights.clone().detach().transpose(-2, -1))
            if len(self.rn.model) > 1:
                internal_layer_mask = self.rn.model[1].mask
                internal_layer_weights = torch.abs(self.rn.model[1].weights.clone().detach().transpose(-2, -1))
            else:
                internal_layer_mask = self.rn.output_layer.mask
                internal_layer_weights = torch.abs(self.rn.output_layer.weights.clone().detach().transpose(-2, -1))

        # initialize class indices if not passed
        if class_indices is None:
            class_indices = torch.arange(self.output_size)
        else:
            class_indices = torch.tensor(class_indices)

        if self.prune_strategy == 'class':
            prune_mask, rewards = self._evaluate_weights(
                input_layer_weights=input_layer_weights,
                class_indices=class_indices,
                mode='identify_important_features'
            )
        elif self.prune_strategy == 'logic':
            # PRUNE RULES BASED ON PERFORMANCE OF EACH RULE
            prune_mask = torch.zeros_like(input_layer_weights.data).bool()
            rewards = torch.zeros_like(input_layer_weights.data)
            all_activations, all_targets, all_sample_idx = self._get_activations(dl)
            prune_mask, rewards = self._evaluate_activations(
                targets=all_targets,
                activations=all_activations,
                sample_idx=all_sample_idx,
                input_layer_weights=input_layer_weights,
                internal_layer_mask=internal_layer_mask,
                mask=prune_mask,
                class_indices=class_indices,
                output_metric=output_metric,
                rewards=rewards,
                dimension=-2,
                mode='identify_important_features',
                objective=objective
            )
        elif self.prune_strategy == 'logic_class':
            prune_mask, rewards = self._evaluate_logic_weights(
                input_layer_weights=input_layer_weights.transpose(-2, -1),
                internal_layer_weights=internal_layer_weights,
                internal_layer_mask=internal_layer_mask,
                class_indices=class_indices,
                mode='identify_important_features'
            )

        # select important features based on mask
        feature_indexes = []
        feature_weights = []
        for i in range(self.output_size):
            feature_indexes += [torch.masked_select(input_layer_mask[i], prune_mask[i]).cpu().numpy().tolist()]
            feature_weights += [torch.masked_select(rewards[i], prune_mask[i]).cpu().numpy().tolist()]

        return feature_indexes, feature_weights

    def _produce_predicate_rewards(
            self,
            feature_indexes: List[int],
            feature_weights: List[float]
    ) -> List[List[float]]:
        """
        Build predicate level rewards for each class.

        Args:
            feature_indexes (List[int]): Important feature indexes.
            feature_weights (List[float]): Important feature weights.

        Returns:
            npt.NDArray: Array of rewards for bandit for each class for all features.
        """
        predicates_total_rewards = []
        for p, v in zip(feature_indexes, feature_weights):
            predicates_rewards = defaultdict(float)
            for predicate, weight in zip(p, v):
                if (not isinstance(predicate, np.int64)
                        and not isinstance(predicate, np.int32)
                        and not isinstance(predicate, int)):
                    if len(predicate) > 1:
                        raise AssertionError("Predicate is not singular!")
                    predicate = predicate[0]
                predicates_rewards[predicate] += abs(weight)
            predicates_total_rewards += [predicates_rewards]

        # rewards for multi-armed bandit
        predicates_rewards = [[p[x] if x in p else 0.0 for x in range(self.input_size)]
                              for p in predicates_total_rewards]

        return predicates_rewards

    def _bayesian_ucb_policy(self, history_df: pd.DataFrame) -> npt.NDArray:
        """
        Apply Bayesian UCB policy to generate feature selection probabilities

        Args:
            history_df (pd.DataFrame): dataframe. Dataset to apply UCB policy to.
        """
        scores = history_df[['predicates', 'rewards']].groupby('predicates').agg(
            {'rewards': ['mean', 'count', 'std']})
        scores.columns = ['mean', 'count', 'std']
        scores['ucb'] = scores['mean'] + (self.ucb_scale * scores['std'] / np.sqrt(scores['count']))
        return softmax(scores.loc[range(self.input_size), 'ucb'].values)

    def _update_policy(
            self,
            feature_indexes: List[int],
            feature_weights: List[float],
            class_indices: List[int] = None
    ):
        """
        Perform policy update for each class.

        Args:
            feature_indexes (List[int]): Important feature indexes.
            feature_weights (List[float]): Important feature weights.
            class_indices (List[int]): Classes for which to perform a policy update.
        """
        predicates_rewards = self._produce_predicate_rewards(feature_indexes, feature_weights)

        if class_indices is None:
            class_indices = range(self.output_size)

        for output in class_indices:
            # features are selected using multi-armed bandit after initial step
            self.predicates_rewards_history[output] = pd.concat([
                self.predicates_rewards_history[output],
                pd.DataFrame({'predicates': range(self.input_size), 'rewards': predicates_rewards[output]}),
            ])
            predicates_probs = self._bayesian_ucb_policy(history_df=self.predicates_rewards_history[output])
            self.predicates_policy[output, :].copy_(torch.tensor(predicates_probs))

    @staticmethod
    def _flip_un_pruned_feature_weights(
            input_layer_mask: torch.Tensor,
            input_layer_weights: torch.Tensor,
            prune_mask: torch.Tensor,
            new_mask: torch.Tensor
    ) -> torch.Tensor:
        """
        Flip weights sign for newly generated features that are in the un-pruned set.  Used to model relations like
        x > feature_value > y.

        Args:
            input_layer_mask (torch.Tensor): Input layer feature mask with feature indexes used in logic.
            input_layer_weights (torch.Tensor): Detached input layer weights from the model.
            prune_mask (torch.Tensor): Boolean mask indicating which positions of the input_layer_mask will be pruned.
            new_mask (torch.Tensor): Newly input layer feature mask.  True positions in prune_mask will be replaced
                with these values

        Returns:
            torch.Tensor: new_input_layer_weights with flipped signs and reinitialized weights
        """
        # determine sign of weights for un-pruned features
        feature_indexes = torch.masked_select(input_layer_mask, ~prune_mask).cpu()
        feature_weights = torch.masked_select(
            input_layer_weights.transpose(-2, -1), ~prune_mask).cpu()
        # if the feature is used in more than one logic, we take the average of the weights
        feature_signs = pd.DataFrame(
            {'feature_weights': feature_weights.numpy(),
             'feature_indexes': feature_indexes.numpy()}
        )
        feature_signs = feature_signs.groupby(
            'feature_indexes')['feature_weights'].mean().to_frame().apply(
            lambda x: np.sign(x)).to_dict()['feature_weights']

        # mask on positions in newly generated feature masks that contain un-pruned features
        new_mask_existing_features_mask = torch.isin(new_mask.cpu(), feature_indexes.cpu())
        # mask on positions that contain un-pruned features and will be pruned this round
        new_mask_existing_features_mask_prune_mask = new_mask_existing_features_mask.cpu() * prune_mask.cpu()
        # reorient to new_mask
        new_mask_existing_features_mask_prune_mask = new_mask_existing_features_mask_prune_mask.transpose(-2, -1)

        # features at each position
        new_mask_weights_orientation = new_mask.clone().transpose(-2, -1)  # clone because we need new_mask later
        new_input_layer_weights = input_layer_weights.clone()  # clone because we need input_layer_weights.size() later

        # view all tensors as vectors
        new_mask_weights_orientation = new_mask_weights_orientation.contiguous().view(-1)
        new_input_layer_weights = new_input_layer_weights.contiguous().view(-1)
        new_mask_existing_features_mask_prune_mask = new_mask_existing_features_mask_prune_mask.contiguous().view(-1)

        # examine each element in vector
        for i, (feature_idx, weight, mask) \
                in enumerate(zip(new_mask_weights_orientation,
                                 new_input_layer_weights,
                                 new_mask_existing_features_mask_prune_mask)):
            # if the element is an un-pruned feature in the newly generated feature mask that will be pruned
            # then flip the sign of the new feature weight
            if mask:
                new_weights_sign = feature_signs[int(feature_idx.cpu().numpy())] * -1
                if new_weights_sign > 0:
                    # new_weight = torch.ones(1).uniform_(0, 0.1)
                    new_weight = torch.ones(1).uniform_(0, 0.2)
                    # new_weight = torch.nn.init.trunc_normal_(torch.ones(1), a=0, b=1)
                    # new_weight = torch.ones(1)
                else:
                    # new_weight = torch.ones(1).uniform_(-0.1, 0)
                    new_weight = torch.ones(1).uniform_(-0.2, 0)
                    # new_weight = torch.nn.init.trunc_normal_(torch.ones(1), a=-1, b=0)
                    # new_weight = torch.ones(1) * -1
                new_input_layer_weights.data[i] = new_weight

        # update the feature weights to use the flipped signs
        new_input_layer_weights = new_input_layer_weights.view(input_layer_weights.size())

        return new_input_layer_weights

    def perform_prune(
            self,
            dl: DataLoader = None,
            class_indices: List[int] = None,
            output_metric: Callable = None,
            objective: str = 'maximize'
    ):
        """
        Prune the network.

        Args:
            dl (DataLoader): PyTorch data loader with training data.
            class_indices (list): List of integers indicating which class policies to update.
            output_metric (sklearn.metrics): Scikit-learn performance metric to use for evaluating logic.
            objective (str): 'maximize' or 'minimize'.  Which direction is the optimal performance for output_metric.
        """
        if self.USE_DATA_PARALLEL:
            input_layer_mask = self.rn.module.model[0].mask
            input_layer_weights = self.rn.module.model[0].weights.clone().detach()
            if len(self.rn.module.model) > 1:
                internal_layer_mask = self.rn.module.model[1].mask
                internal_layer_weights = torch.abs(self.rn.module.model[1].weights.clone().detach().transpose(-2, -1))
            else:
                internal_layer_mask = self.rn.output_layer.mask
                internal_layer_weights = torch.abs(
                    self.rn.module.output_layer.weights.clone().detach().transpose(-2, -1))
        else:
            input_layer_mask = self.rn.model[0].mask
            input_layer_weights = self.rn.model[0].weights.clone().detach()
            if len(self.rn.model) > 1:
                internal_layer_mask = self.rn.model[1].mask
                internal_layer_weights = torch.abs(self.rn.model[1].weights.clone().detach().transpose(-2, -1))
            else:
                internal_layer_mask = self.rn.output_layer.mask
                internal_layer_weights = torch.abs(self.rn.output_layer.weights.clone().detach().transpose(-2, -1))

        # initialize class indices if not passed
        if class_indices is None:
            class_indices = torch.arange(self.output_size)
        else:
            class_indices = torch.tensor(class_indices)

        if self.prune_strategy == 'class':
            prune_mask, _ = self._evaluate_weights(
                input_layer_weights=input_layer_weights,
                class_indices=class_indices,
                mode='prune'
            )
        elif self.prune_strategy == 'logic':
            prune_mask = torch.zeros_like(input_layer_weights.data).bool()
            all_activations, all_targets, all_sample_idx = self._get_activations(dl)
            prune_mask, _ = self._evaluate_activations(
                targets=all_targets,
                activations=all_activations,
                sample_idx=all_sample_idx,
                input_layer_weights=input_layer_weights,
                internal_layer_mask=internal_layer_mask,
                mask=prune_mask,
                class_indices=class_indices,
                output_metric=output_metric,
                rewards=None,
                dimension=-1,
                mode='prune',
                objective=objective
            )
        elif self.prune_strategy == 'logic_class':
            prune_mask, _ = self._evaluate_logic_weights(
                input_layer_weights=input_layer_weights,
                internal_layer_weights=internal_layer_weights,
                internal_layer_mask=internal_layer_mask,
                class_indices=class_indices,
                mode='prune'
            )

        # # don't select used features in new logic
        predicates_policy_to_use = self.predicates_policy.clone()
        for i in range(self.output_size):
            feature_indexes = torch.LongTensor(
                [torch.masked_select(input_layer_mask[i], ~prune_mask[i]).cpu().numpy().tolist()])
            predicates_policy_to_use[i, feature_indexes] /= self.delta

        if not self.add_negations:
            # generate new input mask
            try:
                new_mask = torch.multinomial(
                    predicates_policy_to_use, input_layer_mask.size(1) * input_layer_mask.size(2)).reshape(
                    -1, input_layer_mask.size(1), input_layer_mask.size(2))
            except RuntimeError as e:
                new_mask = torch.multinomial(
                    predicates_policy_to_use, input_layer_mask.size(1) * input_layer_mask.size(2),
                    replacement=True).reshape(
                    -1, input_layer_mask.size(1), input_layer_mask.size(2))

            new_input_layer_weights = self._flip_un_pruned_feature_weights(
                input_layer_mask=input_layer_mask,
                input_layer_weights=input_layer_weights,
                prune_mask=prune_mask,
                new_mask=new_mask
            )
        else:
            # generate new input mask
            try:
                new_mask = torch.multinomial(
                    predicates_policy_to_use, input_layer_mask.size(1) * input_layer_mask.size(2) // 2).reshape(
                    -1, input_layer_mask.size(1), input_layer_mask.size(2) // 2)
                new_mask = torch.cat([new_mask, new_mask.clone()], dim=-1)
            except RuntimeError as e:
                new_mask = torch.multinomial(
                    predicates_policy_to_use, input_layer_mask.size(1) * input_layer_mask.size(2) // 2,
                    replacement=True).reshape(
                    -1, input_layer_mask.size(1), input_layer_mask.size(2) // 2)
                new_mask = torch.cat([new_mask, new_mask.clone()], dim=-1)

            size = (input_layer_weights.size(0), input_layer_weights.size(1) // 2, input_layer_weights.size(2))
            # new_input_layer_weights = torch.ones(size).uniform_(-0.1, 0.1)
            new_input_layer_weights = torch.ones(size).uniform_(-0.2, 0.2)
            # new_input_layer_weights = torch.nn.init.trunc_normal_(torch.ones(size), a=0, b=1)
            # new_input_layer_weights = torch.nn.init.xavier_uniform_(torch.ones(size), gain=torch.nn.init.calculate_gain('relu'))
            # new_input_layer_weights = torch.where(torch.zeros(size).uniform_(0, 1) > 0.5, torch.ones(size), torch.ones(size) * -1)
            new_input_layer_weights = torch.cat([new_input_layer_weights, new_input_layer_weights.clone() * -1], dim=1)

        # update mask
        try:
            input_layer_mask.masked_scatter_(
                prune_mask.to(input_layer_mask.device), new_mask.to(input_layer_mask.device))
        except NotImplementedError as e:
            input_layer_mask = input_layer_mask.to('cpu')
            input_layer_mask.masked_scatter_(
                prune_mask.to(input_layer_mask.device), new_mask.to(input_layer_mask.device))
            input_layer_mask.to(input_layer_weights.device)

        if self.USE_DATA_PARALLEL:
            self.rn.module.model[0].mask.copy_(input_layer_mask)
            self.rn.module.model[0].weights.data.copy_(new_input_layer_weights.to(input_layer_weights.device))
        else:
            self.rn.model[0].mask.copy_(input_layer_mask)
            self.rn.model[0].weights.data.copy_(new_input_layer_weights.to(input_layer_weights.device))

    def update_policy(
            self,
            dl: DataLoader,
            class_indices: List[int] = None,
            output_metric: Callable = None,
            objective='maximize'
    ):
        """
        Update the policy.

        Args:
            dl (DataLoader): PyTorch data loader with training data.
            class_indices (list): List of integers indicating which class policies to update.
            output_metric (sklearn.metrics): Scikit-learn performance metric to use for evaluating logic.
            objective (str): 'maximize' or 'minimize'.  Which direction is the optimal performance for output_metric.
        """
        if self.USE_DATA_PARALLEL:
            feature_indexes, feature_weights = self._identify_important_features(
                dl, class_indices, output_metric, objective)
        else:
            feature_indexes, feature_weights = self._identify_important_features(
                dl, class_indices, output_metric, objective)
        self._update_policy(feature_indexes, feature_weights, class_indices)


__all__ = [BaseBanditNRNModel]



from .blocks import LukasiewiczChannelBlock, VariationalLukasiewiczChannelBlock, AttentionLukasiewiczChannelBlock
from .predicates import BasePredicates
from .utils import BaseConcatenateBlocksLogic

__all__ = [LukasiewiczChannelBlock, BasePredicates, BaseConcatenateBlocksLogic,
           VariationalLukasiewiczChannelBlock, AttentionLukasiewiczChannelBlock]



from .base import BasePredicates
import logging


class Predicates(BasePredicates):

    def __init__(self, feature_names: list):
        """
        Initialize a Predicates object.  The Predicates object is passed to the operands parameter of a
        LukasiewiczLayer or LukasiewiczChannelBlock and enables the explanation functionality of the torchlogic
        blocks, layers and modules.

        Example:
            feature_names = ['feature1', 'feature2', 'feature3', 'feature4']

            input_layer_and = LukasiewiczChannelAndBlock(
                channels=output_size,
                in_features=input_size,
                out_features=layer_sizes,
                n_selected_features=n_selected_features_input,
                parent_weights_dimension='out_features',
                operands=Predicates(feature_names=feature_names)
                )

        Args:
            feature_names (list): A list of feature names.
        """
        super(Predicates, self).__init__(feature_names=feature_names)
        self.logger = logging.getLogger(self.__class__.__name__)



from typing import Tuple
from collections import defaultdict

import numpy as np
import pandas as pd
from sklearn.pipeline import make_pipeline

from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import OneHotEncoder, MinMaxScaler

import torch
from torch.utils.data import Dataset
from torch.utils.data import DataLoader, SubsetRandomSampler
from aix360.algorithms.rbm import FeatureBinarizerFromTrees


class BaseSKLogicEstimator:

    def __init__(
            self,
            binarization: bool = True,
            tree_num: int = 10,
            tree_depth: int = 5,
            tree_feature_selection: float = 0.5,
            thresh_round: int = 3,
            learning_rate: float = 0.1,
            weight_decay: float = 0.001,
            t_0: int = 3,
            t_mult: int = 2,
            epochs: int = 200,
            batch_size: int = 32,
            holdout_pct: float = 0.2,
            early_stopping_plateau_count: int = 20,
            lookahead_steps: int = 0,
            lookahead_steps_size: float = 0.0,
            pin_memory: bool = False,
            persistent_workers: bool = False,
            num_workers: int = 0
    ):
        """
        Base Scikit-Learn torchlogic compatible estimator.

        Args:
            binarization (bool): if true, use feature binarization from trees
            tree_num (int): number of trees for feature binarization from trees
            tree_depth (int): depth of trees for feature binarization from trees
            tree_feature_selection (float): feature selection rate for feature binarization from trees
            thresh_round (int): rounding threshold in decimal places for feature binarzation from trees
            learning_rate (float): learning rate for AdamW optimizer
            weight_decay (float): weight decay for AdamW optimizer
            t_0 (int): T_0 for CosineAnnealingWithWarmRestarts scheduler
            t_mult (int): T_mult for CosineAnnealingWithWarmRestarts scheduler
            epochs (int): epochs for training procedure
            batch_size (int): batch size for data loading
            holdout_pct (float): percentage of training data used as holdout for early stopping
            early_stopping_plateau_count (int): number of epochs without improvement for early stopping
            lookahead_steps (int): number of steps for lookahead optimization. If zero, lookahead not used.
            lookahead_steps_size (float): step size for lookahead optimization.  If lookahead_steps is zero not used.
            pin_memory (bool): pin_memory for data loaders.
            persistent_workers (bool): persistent workers for data loaders.
            num_workers (int): number of workers for data loaders.
        """
        # add hyper-parameters
        self.binarization = binarization
        self.tree_num = tree_num
        self.tree_depth = tree_depth
        self.tree_feature_selection = tree_feature_selection
        self.thresh_round = thresh_round
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.t_0 = t_0
        self.t_mult = t_mult
        self.epochs = epochs
        self.batch_size = batch_size
        self.holdout_pct = holdout_pct
        self.early_stopping_plateau_count = early_stopping_plateau_count
        self.lookahead_steps = lookahead_steps
        self.lookahead_steps_size = lookahead_steps_size
        self.pin_memory = pin_memory
        self.persistent_workers = persistent_workers
        self.num_workers = num_workers

        self.feature_names = None

        # data encoding
        self.mixed_pipe = None
        self.numeric_features = None
        self.categorical_features = None
        self._fbt_is_fitted = False
        self.min_max_features_dict = None

        # create feature binarizer
        if self.binarization:
            self.fbt = FeatureBinarizerFromTrees(
                treeNum=self.tree_num,
                treeDepth=self.tree_depth,
                treeFeatureSelection=self.tree_feature_selection,
                threshRound=self.thresh_round,
                randomState=0
            )

        self.model = None

    def _create_holdout_samplers(
            self,
            train_dataset: Dataset
    ) -> Tuple[SubsetRandomSampler, SubsetRandomSampler]:
        """
        Create holdout data sampler for training.

        Args:
            train_dataset (Dataset): the training dataset

        Returns:
            Tuple[SubsetRandomSampler, SubsetRandomSampler]: training data sampler, holdout data sampler
        """
        assert self.holdout_pct > 0 and self.holdout_pct < 1, "holdout_pct must be between 0 and 1"
        g = torch.Generator()
        g.manual_seed(0)

        train_size = len(train_dataset)
        indices = list(range(train_size))
        np.random.seed(0)
        np.random.shuffle(indices)

        train_holdout_split_index = int(np.floor(self.holdout_pct * train_size))
        train_idx, train_holdout_idx = indices[train_holdout_split_index:], indices[:train_holdout_split_index]

        train_sampler = SubsetRandomSampler(train_idx)
        train_holdout_sampler = SubsetRandomSampler(train_holdout_idx)

        return train_sampler, train_holdout_sampler

    def _generate_training_data_loaders(self, dataset: Dataset) -> Tuple[DataLoader, DataLoader]:
        """
        Generate the training data loaders.

        Args:
            dataset (Dataset): the training data pytorch dataset

        Returns:
            Tuple[DataLoader, DataLoader]: the training data loader
        """
        g = torch.Generator()
        g.manual_seed(0)

        train_sampler, train_holdout_sampler = self._create_holdout_samplers(dataset)

        train_dataloader = DataLoader(
            dataset, batch_size=self.batch_size, generator=g, sampler=train_sampler,
            pin_memory=self.pin_memory, persistent_workers=self.persistent_workers, num_workers=self.num_workers
            # very important to optimize these settings in production
        )
        train_holdout_dataloader = DataLoader(
            dataset, batch_size=self.batch_size, generator=g, sampler=train_holdout_sampler,
            pin_memory=self.pin_memory, persistent_workers=self.persistent_workers, num_workers=self.num_workers
            # very important to optimize these settings in production
        )

        return train_dataloader, train_holdout_dataloader

    def _handle_non_binarized_categorical_feature_names(self, X: pd.DataFrame):
        if not self.binarization:
            new_cols = []
            for col in X.columns:
                if col.find("was_") > -1:
                    col = col.replace("was_", " ") + " was"
                new_cols += [col]
            X.columns = new_cols
        return X

    def _fit_transform_encode_data(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Fit and transform the features for the model.

        Args:
            X (pd.DataFrame): DataFrame of features

        Returns:
            pd.DataFrame: Transformed DataFrame of features
        """
        self.numeric_features = X.select_dtypes(include='number')
        self.categorical_features = X.select_dtypes(exclude='number')
        numeric_feature_names = list(self.numeric_features.columns)
        categorical_feature_names = list(self.categorical_features.columns)

        if not self.binarization:
            mixed_encoded_preprocessor = ColumnTransformer(
                [
                    (
                        "numeric",
                        MinMaxScaler(),
                        numeric_feature_names
                    ),
                    (
                        "categorical",
                        OneHotEncoder(handle_unknown="ignore", sparse_output=False, dtype=np.bool_),
                        categorical_feature_names,
                    )
                ],
                verbose_feature_names_out=False,
            )
        else:
            mixed_encoded_preprocessor = ColumnTransformer(
                [
                    (
                        "numeric",
                        "passthrough",
                        numeric_feature_names
                    ),
                    (
                        "categorical",
                        OneHotEncoder(handle_unknown="ignore", sparse_output=False, dtype=np.bool_),
                        categorical_feature_names,
                    )
                ],
                verbose_feature_names_out=False,
            )

        mixed_encoded_preprocessor.set_output(transform='pandas')

        self.mixed_pipe = make_pipeline(
            mixed_encoded_preprocessor,
        )

        X = self.mixed_pipe.fit_transform(X)
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)

        if not self.binarization:
            self.numeric_features = X[numeric_feature_names].copy()

        X = self._handle_non_binarized_categorical_feature_names(X)

        return X

    def _transform_encode_data(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Apply fitted transformation of features to new data.

        Args:
            X (pd.DataFrame): DataFrame of features.

        Returns:
            pd.DataFrame: Transformed DataFrame of features
        """
        X = self.mixed_pipe.transform(X)
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)

        X = self._handle_non_binarized_categorical_feature_names(X)

        return X

    @staticmethod
    def _transform_column_names(df: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the column names of encoded features for use in explanations.

        Args:
            df (pd.DataFrame): DataFrame of transformed data.

        Returns:
            pd.DataFrame: DataFrame of transformed data with column names needed for explanations.
        """
        df.columns = list(map(
            lambda x: str(x).replace("(", "").replace(")", "").replace(" ", "_")
            .replace("/", "_").replace(",", "").replace("'", "")
            .replace("_", " ").replace(">=", "greater than or equal to")
            .replace("<=", "less than or equal to").replace(">", "greater than")
            .replace("<", "less than").replace("cat  ", "")
            .replace("num  ", ""),
            df.columns.to_flat_index()))
        return df

    def initialize_binarizer(self, fitted_binarizer: FeatureBinarizerFromTrees) -> None:
        """
        Initialize the feature binarizer from a previously fitted binarizer.

        Args:
            fitted_binarizer (FeatureBinarizerFromTrees): Fitted FeatureBinarizerFromTrees object

        Returns:
            None
        """
        self.fbt = fitted_binarizer
        self._fbt_is_fitted = True

    def _fit_transform_binarize_features(self, X: pd.DataFrame, y: pd.DataFrame) -> pd.DataFrame:
        """
        Fit and transform input data with FeatureBinarizationFromTrees

        Args:
            X (pd.DataFrame): DataFrame of transformed features.
            y (pd.DataFrame): DataFrame of targets.

        Returns:
            pd.DataFrame: Binarized features.
        """
        X_numeric = X.select_dtypes(include='number')
        X_categorical = X.select_dtypes(exclude='number')
        numeric_columns = list(X_numeric.columns)
        categorical_columns = list(X_categorical.columns)

        if not self._fbt_is_fitted:
            X_numeric = self.fbt.fit_transform(X_numeric, y)
        else:
            X_numeric = self.fbt.transform(X_numeric)
        X_numeric = self._transform_column_names(X_numeric)

        feature_names = list(X_numeric.columns) + list(categorical_columns)
        X = np.hstack([X_numeric.values, X_categorical.values])

        X = pd.DataFrame(X, columns=feature_names)
        X.rename(columns=lambda x: x.replace("_", " "), inplace=True)
        self._fbt_is_fitted = True

        return X

    def _transform_binarize_features(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform input features with fitted FeatureBinarizationFromTrees.

        Args:
            X (pd.DataFrame): Un-binarized features.

        Returns:
            pd.DataFrame: Binarized features.
        """
        X_numeric = X.select_dtypes(include='number')
        X_categorical = X.select_dtypes(exclude='number')
        numeric_columns = list(X_numeric.columns)
        categorical_columns = list(X_categorical.columns)

        X_numeric = self.fbt.transform(X_numeric)
        X_numeric = self._transform_column_names(X_numeric)

        feature_names = list(X_numeric.columns) + list(categorical_columns)
        X = np.hstack([X_numeric.values, X_categorical.values])

        X = pd.DataFrame(X, columns=feature_names)
        X.rename(columns=lambda x: x.replace("_", " "), inplace=True)

        return X

    def _encode_prediction_data(self, X):
        # ecode data in necessary format
        X = self._transform_encode_data(X)
        if self.binarization:
            X = self._transform_binarize_features(X)
        return X

    @staticmethod
    def _handle_non_dataframe_features(X):
        X = pd.DataFrame(X, columns=[f'feature {i}' for i in range(X.shape[1])])
        for col in X.columns:
            try:
                X[col] = X[col].astype(float).copy()
            except ValueError:
                pass
        return X

    @staticmethod
    def _handle_non_dataframe_targets(y):
        y = pd.DataFrame(y, columns=[f'Class {i}' for i in range(y.shape[1])])
        y = y.apply(lambda x: x.astype(np.float32), axis=1)
        return y

    def _handle_empty_feature_names(self, X: pd.DataFrame):
        # create textual feature names if they are not given by user
        if self.feature_names is None:
            self.feature_names = [f"the {x} was" for x in X.columns]
            X.columns = self.feature_names
        else:
            X.columns = self.feature_names
        return X

    def get_base_params(self, deep=False) -> dict:
        """
        Get parameters of the model

        Args:
            deep (bool): unused parameter

        Returns:
            dict: dictionary of parameters
        """
        return {
            'binarization': self.binarization,
            'tree_num': self.tree_num,
            'tree_depth': self.tree_depth,
            'tree_feature_selection': self.tree_feature_selection,
            'thresh_round': self.thresh_round,
            'learning_rate': self.learning_rate,
            'weight_decay': self.weight_decay,
            't_0': self.t_0,
            't_mult': self.t_mult,
            'epochs': self.epochs,
            'batch_size': self.batch_size,
            'holdout_pct': self.holdout_pct,
            'early_stopping_plateau_count': self.early_stopping_plateau_count,
            'lookahead_steps': self.lookahead_steps,
            'lookahead_steps_size': self.lookahead_steps_size,
            'pin_memory': self.pin_memory,
            'persistent_workers': self.persistent_workers,
            'num_workers': self.num_workers
        }

    def _get_instance_params(self):
        """Base class has no instance params"""
        return {}

    def get_params(self, deep=False):
        """
        Get parameters of the model

        Args:
            deep (bool): unused parameter

        Returns:
            dict: dictionary of parameters
        """
        instance_params = self._get_instance_params()
        base_params = self.get_base_params()
        base_params.update(instance_params)

        return base_params

    def _get_param_names(self):
        """
        Get the parameter names.

        Returns:
            list: list of parameter names
        """
        return list(self.get_params().keys())

    def set_params(self, **params):
        """Set the parameters of this estimator.

        The method works on simple estimators as well as on nested objects
        (such as :class:`~sklearn.pipeline.Pipeline`). The latter have
        parameters of the form ``<component>__<parameter>`` so that it's
        possible to update each component of a nested object.

        Parameters
        ----------
        **params : dict
            Estimator parameters.

        Returns
        -------
        self : estimator instance
            Estimator instance.
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params()

        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition("__")
            if key not in valid_params:
                local_valid_params = self._get_param_names()
                raise ValueError(
                    f"Invalid parameter {key!r} for estimator {self}. "
                    f"Valid parameters are: {local_valid_params!r}."
                )

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)

        return self



import numpy as np

## Placeholder for the Maximial Information Coefficient (MIC) estimation implementation from Minepy, which is not available in python >=3.11

# A minimal approximation of Maximial Information Coeff estimation (Reshef, 2011) https://pmc.ncbi.nlm.nih.gov/articles/PMC3325791/
# Only dependent on numpy. Slow as dirt in testing (~20x slower) with about 20% MAPE in comparison to the canonical package estimation time and estimates.

## Pseudo-code:
# https://pmc.ncbi.nlm.nih.gov/articles/instance/3325791/bin/NIHMS358982-supplement-Supplemental_Figures_and_Tables.pdf


#### WISH LIST ####
# - Adaptive binning based on data distribution (e.g., quantiles) (Can also try equi-depth binning)
# - More efficient mutual information calculation (currently uses nested loops)
# - Math review for correctness, especially in normalization, binning, and parameter choices
# - More verbose error handling and logging

def bin_series(x, bins):
    """Assigns each value to a bin index."""
    # Convert input to numpy array and flatten to 1D
    x = np.asarray(x).flatten()
    
    # Handle edge case of empty array
    if len(x) == 0:
        return np.array([], dtype=int)
    
    # Handle edge case where all values are identical (zero variance)
    if np.var(x) == 0:
        return np.zeros(len(x), dtype=int)
    
    # Check if requested bins is less than or equal to unique values
    if bins <= len(np.unique(x)):
        # Use numpy's histogram to create bin edges
        _, edges = np.histogram(x, bins=bins)
        # Digitize assigns each value to a bin, subtract 1 for 0-based indexing
        binned = np.digitize(x, edges[:-1], right=False) - 1
        # Clip values to ensure they're within valid bin range
        binned = np.clip(binned, 0, bins - 1)
        return binned
    else:
        # If more bins requested than unique values, use unique value mapping
        unique_vals = np.unique(x)
        return np.searchsorted(unique_vals, x)


def mutual_information_2d(x_binned, y_binned):

    try:
        # Ensure inputs are integer arrays for indexing
        x_binned = np.asarray(x_binned, dtype=int)
        y_binned = np.asarray(y_binned, dtype=int)
        
        #  Determine the size of the joint histogram matrix
        max_x = int(np.max(x_binned)) + 1
        max_y = int(np.max(y_binned)) + 1
        
        # Initialize joint histogram matrix with zeros
        joint_hist = np.zeros((max_x, max_y))
        
        # Populate joint histogram by counting co-occurrences
        for i in range(len(x_binned)):
            joint_hist[x_binned[i], y_binned[i]] += 1
        
        # Convert counts to probabilities
        n = len(x_binned)
        joint_prob = joint_hist / n
        
        # Calculate marginal probabilities
        px = np.sum(joint_prob, axis=1)  # Sum over y-axis to get P(X)
        py = np.sum(joint_prob, axis=0)  # Sum over x-axis to get P(Y)
        
        ####### These nested loops are very inefficient. Could vectorize this calculation? ########
        # Calculate mutual information using the formula: MI = sum(P(x,y) * log(P(x,y) / (P(x) * P(y))))
        mi = 0.0
        for i in range(max_x):
            for j in range(max_y):
                if joint_prob[i, j] > 0 and px[i] > 0 and py[j] > 0:
                    mi += joint_prob[i, j] * np.log(joint_prob[i, j] / (px[i] * py[j]))
        
        # Ensure MI is non-negative (theoretically) it should be, but you never know for sure)
        return max(mi, 0.0)
    except Exception as e:
        # Generic exception handling - prints error and returns 0
        print(f"MI calculation error: {e}")
        return 0.0
    
def compute_mic(x, y, alpha=0.6, c=5, max_bins=None):
    """Compute the Maximial Information Coefficient (MIC) between two variables x and y"""
    
    # Convert inputs to flattened numpy arrays
    x = np.asarray(x).flatten()
    y = np.asarray(y).flatten()
    
    # Handle edge case of (x,y) length mismatch
    if len(x) != len(y):
        raise ValueError(f"x and y must have same length: {len(x)} vs {len(y)}")
    
    # Handle edge case of very small datasets
    if len(x) < 4:
        return 0.0, (0, 0)
    
    # Calculate sample size
    n = len(x)

    #### Alpha scaling logic seems questionable ####
    # Calculate Maximum number of bins based on sample size and alpha
    B = min(int(n ** alpha), 500) if alpha <= 1 else min(int(n ** (alpha / 100.0)), 500)
    max_bins = min(int(np.sqrt(B)), 25) if max_bins is None else max_bins
    
    max_mic = 0.0
    best_grid = (0, 0)

    # Iterate through all possible grid sizes
    for x_bins in range(2, max_bins + 1):
        for y_bins in range(2, max_bins + 1):
            
            # Skip if grid size exceeds B parameter
            if x_bins * y_bins >= B:
                continue

            # This is a heuristic to avoid extreme binning ratios
            if max(x_bins, y_bins) > c * min(x_bins, y_bins):
                continue
            
            # Skip if binning resulted in fewer than 2 unique bins
            try:
                # Bin the data using the current grid size
                x_binned = bin_series(x, x_bins)
                y_binned = bin_series(y, y_bins)
                if len(np.unique(x_binned)) < 2 or len(np.unique(y_binned)) < 2:
                    continue
                
                # Calculate MIC score and update best if it's better
                mi = mutual_information_2d(x_binned, y_binned)
                
                # Normalize MI by log of minimum grid dimension
           ###### I am not very confident this normalization is corrrect ######
                norm = np.log(min(x_bins, y_bins))
                if norm > 0 and mi > 0:
                    mic = mi / norm
                    if mic > max_mic:
                        max_mic = mic
                        best_grid = (x_bins, y_bins)
            
            # Skip this grid size if any error occurs
            except Exception:
                continue
    
    # Return the best MIC score and corresponding grid size
    return max_mic, best_grid


## I asked claude to wrap everything up in a single function so we can run all cov at once, and output the desired matrix
## Implementation is a bit verbose, but it handles both single covariate MIC and pairwise (X) MIC matrix computation.
def compute_mic_matrix(X, y=None, alpha=0.6, c=5, max_bins=None):
    """
    Compute MIC for multiple covariates.
    
    Parameters:
    -----------
    X : array-like, shape (n_samples, n_features) or (n_samples,)
        Input covariates. If 2D, each column is a covariate.
    y : array-like, shape (n_samples,), optional
        Target variable. If None, computes pairwise MIC between all covariates in X.
    alpha, c, max_bins : parameters passed to compute_mic_e
    
    Returns:
    --------
    If y is provided:
        - mic_scores : array, shape (n_features,)
        - best_grids : list of tuples, shape (n_features,)
    
    If y is None:
        - mic_matrix : array, shape (n_features, n_features)
        - best_grids : list of lists of tuples
    """
    X = np.asarray(X)
    
    # Handle 1D input
    if X.ndim == 1:
        X = X.reshape(-1, 1)
    
    n_samples, n_features = X.shape
    
    # Not really needed, but claude included handling for no-target case.
    if y is not None:
        # Compute MIC between each covariate and target y
        y = np.asarray(y).flatten()
        if len(y) != n_samples:
            raise ValueError(f"X and y must have same number of samples: {n_samples} vs {len(y)}")
        
        mic_scores = np.zeros(n_features)
        best_grids = []
        
        for i in range(n_features):
            mic_score, best_grid = compute_mic(X[:, i], y, alpha, c, max_bins)
            mic_scores[i] = mic_score
            best_grids.append(best_grid)
        
        return mic_scores, best_grids
    
    else:
        # Compute pairwise MIC matrix between all covariates
        mic_matrix = np.zeros((n_features, n_features))
        best_grids = [[None for _ in range(n_features)] for _ in range(n_features)]
        
        for i in range(n_features):
            for j in range(i, n_features):
                if i == j:
                    mic_matrix[i, j] = 1.0  # Perfect correlation with itself
                    best_grids[i][j] = (0, 0)
                else:
                    mic_score, best_grid = compute_mic(X[:, i], X[:, j], alpha, c, max_bins)
                    mic_matrix[i, j] = mic_score
                    mic_matrix[j, i] = mic_score  # Symmetric
                    best_grids[i][j] = best_grid
                    best_grids[j][i] = best_grid
        
        return mic_matrix, best_grids


import logging
from typing import Tuple

import numpy as np
import pandas as pd
from scipy.special import expit

import torch
from torch.utils.data import DataLoader

from nrn.utils.explanations import register_hooks
from ._base_mixin import BaseReasoningNetworkMixin
from nrn.utils.explanations import simplification


class ReasoningNetworkClassifierMixin(BaseReasoningNetworkMixin):

    ROUNDING_PRECISION = 32
    EPS = 1e-2

    def __init__(
            self,
            output_size: int,
            logits: bool = False
    ):
        super(ReasoningNetworkClassifierMixin, self).__init__()
        self.output_size = output_size
        self.logits = logits
        self.logger = logging.getLogger(self.__class__.__name__)

    def explain(
            self,
            quantile: float = 0.5,
            required_output_thresholds: torch.Tensor = torch.tensor(0.9),
            threshold: float = None,
            explain_type: str = 'both',
            print_type: str = 'logical',
            target_names: list = ['positive'],
            explanation_prefix: str = "A sample is in the",
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            decision_boundary: float = 0.5,
            show_bounds: bool = True,
            simplify: bool = False,
            exclusions: list[str] = None
    ) -> str:
        """
        Produce a global explanation.

        Args:
            quantile (float): Quantile of logic to produce.
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            target_names (list): list of target names.
            explanation_prefix (str): text to precede explanation.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn Transformation): Skleran inverse transformation function
            decision_boundary (float): Classification decision boundary
            show_bounds (bool): include numeric boundary used in logic
            simplify (bool): If True, return a simplified version of the explanation using logical rules

        Returns:
            str: explanation
        """
        if not isinstance(required_output_thresholds, torch.Tensor):
            required_output_thresholds = torch.tensor(required_output_thresholds).float()

        # binary classification?
        if self.output_size == 1:
            force_negate = required_output_thresholds < decision_boundary
            force_negate_str = f"not " if force_negate else ""
        else:
            force_negate = False
            force_negate_str = ""

        explanation = []
        for output in range(self.output_size):
            op_explain = self.root_layer.explain(
                quantile=quantile,
                required_output_thresholds=required_output_thresholds,
                threshold=threshold,
                parent_weights=torch.ones(1),
                parent_logic_type='Or',
                depth=0,
                explain_type=explain_type,
                print_type=print_type if print_type == 'natural' else 'logical',
                channel=output,
                force_negate=force_negate,
                ignore_uninformative=ignore_uninformative,
                rounding_precision=self.ROUNDING_PRECISION,
                inverse_transform=inverse_transform,
                show_bounds=show_bounds,
                original_rounding_precision=rounding_precision,
            )
            explanation_str = ', '.join(np.unique(op_explain).tolist())

            if not explanation_str:
                UserWarning("Could not generate explanation with given settings.  "
                            "Searching setting space instead.")

                search_params = self._build_search_space(
                    quantile=quantile,
                    explain_type=explain_type,
                    ignore_uninformative=ignore_uninformative,
                    rounding_precision=rounding_precision
                )
                for params in search_params:
                    op_explain = self.root_layer.explain(
                        required_output_thresholds=required_output_thresholds,
                        threshold=threshold,
                        parent_weights=torch.ones(1),
                        parent_logic_type='Or',
                        depth=0,
                        print_type=print_type if print_type == 'natural' else 'logical',
                        channel=output,
                        force_negate=force_negate,
                        inverse_transform=inverse_transform,
                        show_bounds=show_bounds,
                        original_rounding_precision=rounding_precision,
                        **params
                    )
                    explanation_str = ', '.join(np.unique(op_explain).tolist())
                    if explanation_str:
                        break

            if not explanation_str:
                raise RuntimeError("Could not specify parameters to create explanation.  "
                                   "Check that required_output_thresholds is in range of model outputs.")

            explanation_str = simplification(
                explanation_str,
                print_type,
                simplify,
                sample_level=False,
                ndigits=rounding_precision,
                exclusions=exclusions
            )
            explanation += [f"{explanation_prefix} "
                            f"{force_negate_str}{target_names[output]} because: \n\n{explanation_str}"]

        explanation = '\n\n'.join(explanation)
        if print_type == 'natural':
            explanation = '.  '.join([x[:1].capitalize() + x[1:] for x in explanation.split('.  ')])

        return explanation

    def explain_samples(
            self,
            x: torch.Tensor,
            quantile: float = 0.5,
            threshold: float = None,
            target_names: list = ['positive'],
            explain_type: str = 'both',
            print_type: str = 'logical',
            sample_explanation_prefix: str = "Sample was in the",
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            decision_boundary: float = None,
            show_bounds: bool = True,
            simplify: bool = False,
            exclusions: list[str] = None,
            min_max_feature_dict: dict = None,
            feature_importances: bool = False,
            feature_importances_type: str = 'weight'
    ) -> str:
        """
        Produce a sample explanation.

        Args:
            x (torch.Tensor): features for sample.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            target_names (list): list of target names.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            sample_explanation_prefix (str): text to precede explanation.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn Transformation): Skleran inverse transformation function
            decision_boundary (float): Classification decision boundary
            show_bounds (bool): include numeric boundary used in logic
            simplify (bool): If True, return a simplified version of the explanation using logical rules

        Returns:
            str: explanation
        """
        x = x.to(self.root_layer.weights.device)
        sample_explanations = []
        for i in range(x.size(0)):

            # collect predictions and outputs from each layer
            outputs = {}
            register_hooks(self.rn, outputs)
            predictions = self.rn(x[i].unsqueeze(0))

            # boosted predictions
            boosted_predictions = None
            if hasattr(self, 'xgb'):
                if self.xgb_is_fitted:
                    boosted_predictions = predictions.cpu().detach() + self.xgb.predict(
                        x[i].unsqueeze(0).detach().cpu())

            if self.logits:
                predictions = torch.exp(predictions)/(torch.exp(predictions) + 1)

            # collect explanations
            explanation = []
            if decision_boundary is None and self.output_size > 1:
                if boosted_predictions is None:
                    class_predictions = [predictions[0].detach().cpu().argmax()]
                else:
                    class_predictions = [boosted_predictions[0].argmax()]
                required_output_threshold = [predictions[0, class_predictions[0]].detach().cpu().squeeze()]
            elif decision_boundary is not None and self.output_size > 1:
                if boosted_predictions is None:
                    required_output_threshold = [predictions[0, j].detach().cpu().squeeze()
                                                 for j in range(self.output_size)
                                                 if predictions[0, j] > decision_boundary]
                    np_predictions = predictions[0].detach().cpu().squeeze()
                else:
                    required_output_threshold = [predictions[0, j].detach().cpu().squeeze()
                                                 for j in range(self.output_size)
                                                 if boosted_predictions[0, j] > decision_boundary]
                    np_predictions = boosted_predictions[0].squeeze()
                class_predictions = torch.where(np_predictions > decision_boundary)[0]

            # binary classification?
            if self.output_size == 1:
                if boosted_predictions is None:
                    class_predictions = [predictions[0].detach().cpu().argmax()]
                    negation_threshold = [predictions[0, class_predictions[0]].detach().cpu().squeeze()]
                else:
                    class_predictions = [boosted_predictions[0].argmax()]
                    negation_threshold = [boosted_predictions[0, class_predictions[0]].squeeze()]
                required_output_threshold = [predictions[0, class_predictions[0]].detach().cpu().squeeze()]
                decision_boundary = 0.5 if decision_boundary is None else decision_boundary
                force_negate = negation_threshold[0] < decision_boundary
                force_negate_str = f"not " if force_negate else ""
                parent_weights = torch.ones(1)
            else:
                force_negate = False
                force_negate_str = ""
                parent_weights = torch.ones(1)

            if feature_importances:
                assert len(class_predictions) < 2, "Feature importances only works for binary classificaiton problems"

            for rot, class_prediction in zip(required_output_threshold, class_predictions):
                op_explain = self.root_layer.explain_sample(
                    outputs_dict=outputs,
                    # the negative class is asking the question "why didn't you get a higher value?"
                    # also solves for numerical instability.
                    required_output_thresholds=min(rot * (1 + self.EPS if force_negate else 1 - self.EPS), 1.0),
                    quantile=quantile,
                    threshold=threshold,
                    parent_weights=parent_weights,
                    parent_logic_type='And',
                    depth=0,
                    explain_type=explain_type,
                    print_type=print_type if print_type == 'natural' else 'logical',
                    input_features=x[i].cpu(),
                    channel=class_prediction,
                    force_negate=force_negate,
                    ignore_uninformative=ignore_uninformative,
                    rounding_precision=self.ROUNDING_PRECISION,
                    inverse_transform=inverse_transform,
                    show_bounds=show_bounds,
                    original_rounding_precision=rounding_precision,
                    feature_importances=feature_importances,
                    feature_importances_type=feature_importances_type
                )
                explanation_str = ', '.join(np.unique(op_explain).tolist())

                if feature_importances and not explanation_str:
                    raise RuntimeError("Could not produce explanation for feature importances! Check print_samples output for details.")

                if not explanation_str:
                    UserWarning("Could not generate explanation with given settings.  "
                                "Searching setting space instead.")

                    search_params = self._build_search_space(
                        quantile=quantile,
                        explain_type=explain_type,
                        ignore_uninformative=ignore_uninformative,
                        rounding_precision=rounding_precision
                    )
                    for params in search_params:
                        op_explain = self.root_layer.explain_sample(
                            outputs_dict=outputs,
                            # the negative class is asking the question "why didn't you get a higher value?"
                            # also solves for numerical instability.
                            required_output_thresholds=min(rot * (1 + self.EPS if force_negate else 1 - self.EPS), 1.0),
                            threshold=threshold,
                            parent_weights=parent_weights,
                            parent_logic_type='And',
                            depth=0,
                            print_type=print_type if print_type == 'natural' else 'logical',
                            input_features=x[i].cpu(),
                            channel=class_prediction,
                            force_negate=force_negate,
                            inverse_transform=inverse_transform,
                            show_bounds=show_bounds,
                            original_rounding_precision=rounding_precision,
                            feature_importance=feature_importances,
                            feature_importances_type=feature_importances_type
                            **params
                        )
                        explanation_str = ', '.join(np.unique(op_explain).tolist())
                        if explanation_str:
                            break

                if not explanation_str:
                    raise RuntimeError("Could not produce explanation! Check print_samples output for details.")

                if print_type in ['logical', 'logical-natural'] or feature_importances:
                    # TODO: fix simplification and feature importances for the multi label case
                    explanation_tree = simplification(
                        explanation_str,
                        print_type,
                        simplify=simplify,
                        sample_level=True,
                        ndigits=rounding_precision,
                        exclusions=exclusions,
                        min_max_feature_dict=min_max_feature_dict,
                        feature_importances=feature_importances,
                        verbose=False # TODO use False when done testing
                    )
                    explanation_str = explanation_tree.tree_to_string()

                    if feature_importances:
                        importances_dict = {}
                        self._feature_importances = explanation_tree.get_feature_importances(importances_dict=importances_dict)

                explanation += [
                    f"{i}: {sample_explanation_prefix} "
                    f"{force_negate_str}{target_names[class_prediction]} because: \n\n{explanation_str}"]

                sample_explanations += ['\n'.join(explanation)]
        sample_explanations = [x for x in sample_explanations if x]

        sample_explanations = '\n\n'.join(sample_explanations)
        if print_type == 'natural':
            sample_explanations = '.  '.join([x[:1].capitalize() + x[1:] for x in sample_explanations.split('.  ')])

        return sample_explanations

    def get_feature_importances(
            self,
            x: torch.Tensor,
            quantile: float = 1.,
            decision_boundary = None,
            feature_importances_type: str = ''
            ):
        _ = self.explain_samples(
            x,
            quantile=1. if feature_importances_type in ['weight', 'weight_proportion'] else quantile,
            threshold=None,
            target_names=[f'class_{i}' for i in range(self.output_size)],
            explain_type='both',
            print_type='logical',
            sample_explanation_prefix="Sample was in the",
            ignore_uninformative=False,
            rounding_precision=3,
            inverse_transform=None,
            decision_boundary=0. if feature_importances_type in ['weight', 'weight_proportion'] else decision_boundary,
            show_bounds=False,
            use_llm=False,
            simplify=False,
            exclusions=None,
            min_max_feature_dict=None,
            feature_importances=True,
            feature_importances_type=feature_importances_type)
        feature_importances_values = {key: sum(value) for key, value in self._feature_importances.items()}
        del self._feature_importances
        self._feature_importances = dict()
        return feature_importances_values

    @staticmethod
    def aggregate_feature_importances(feat_imp: dict):
        print(feat_imp)
        agg_feat_imp = dict()
        for key, value in feat_imp.items():
            if not " was" in key:
                continue
            agg_feat = key.split(" was")[0]
            if not agg_feat in agg_feat_imp:
                agg_feat_imp[agg_feat] = value
            else:
                agg_feat_imp[agg_feat] += value
        return agg_feat_imp

    def print(
            self,
            quantile=0.5,
            required_output_thresholds: torch.Tensor = None,
            threshold=None,
            explain_type='both',
            print_type='logical',
            target_names=['positive'],
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            decision_boundary: float = 0.5,
            show_bounds: bool = True
    ):
        """
        Produce a global explanation view of the model.

        Args:
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            target_names (list): list of target names.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn Transformation): Skleran inverse transformation function
            decision_boundary (float): Classification decision boundary
            show_bounds (bool): include numeric boundary used in logic
        """
        # binary classification?
        if self.output_size == 1:
            force_negate = required_output_thresholds < decision_boundary
            force_negate_str = f"negative of " if force_negate else ""
        else:
            force_negate = False
            force_negate_str = ""

        for output in range(self.output_size):
            print(f"REASONING NETWORK MODEL FOR: {force_negate_str}{target_names[output]}")
            self.root_layer.print(
                quantile=quantile,
                required_output_thresholds=required_output_thresholds,
                threshold=threshold,
                parent_weights=torch.ones(1),
                parent_logic_type='Or',
                depth=0,
                explain_type=explain_type,
                print_type=print_type,
                channel=output,
                force_negate=force_negate,
                ignore_uninformative=ignore_uninformative,
                rounding_precision=rounding_precision,
                inverse_transform=inverse_transform,
                show_bounds=show_bounds
            )

    def print_samples(
            self,
            x: torch.Tensor,
            quantile: float = 0.5,
            threshold: float = None,
            target_names: list = ['positive'],
            explain_type: str = 'both',
            print_type: str = 'logical',
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            decision_boundary: float = 0.5,
            show_bounds: bool = True
    ):
        """
        Print a sample explanation view of the model.

        Args:
            x (torch.Tensor): features for sample.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            target_names (list): list of target names.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn Transformation): Skleran inverse transformation function
            decision_boundary (float): Classification decision boundary
            show_bounds (bool): include numeric boundary used in logic
        """
        x = x.to(self.root_layer.weights.device)
        for i in range(x.size(0)):

            # collect predictions and outputs from each layer
            outputs = {}
            register_hooks(self.rn, outputs)
            predictions = self.rn(x[i].unsqueeze(0))

            # boosted predictions
            boosted_predictions = None
            if hasattr(self, 'xgb'):
                if self.xgb_is_fitted:
                    boosted_predictions = predictions.cpu().detach() + self.xgb.predict(
                        x[i].unsqueeze(0).detach().cpu())

            if self.logits:
                predictions = torch.exp(predictions)/(torch.exp(predictions) + 1)

            # collect explanations
            parent_weights = torch.ones(1)
            if boosted_predictions is None:
                class_prediction = predictions[0].detach().cpu().argmax()
                negation_threshold = predictions[0, class_prediction].detach().cpu().squeeze()
            else:
                class_prediction = boosted_predictions[0].argmax()
                negation_threshold = boosted_predictions[0, class_prediction].squeeze()
            required_output_threshold = predictions[0, class_prediction].detach().cpu().squeeze()

            # binary classification?
            if self.output_size == 1:
                force_negate = negation_threshold < decision_boundary
                force_negate_str = f"negative of " if force_negate else ""
            else:
                force_negate = False
                force_negate_str = ""

            print(f"REASONING NETWORK MODEL FOR: {force_negate_str}{target_names[class_prediction]}")
            self.root_layer.print_sample(
                outputs_dict=outputs,
                # the negative class is asking the question "why didn't you get a higher value?"
                # also solves for numerical instability.
                required_output_thresholds=min(
                    required_output_threshold * (1 + self.EPS if force_negate else 1 - self.EPS), 1.0),
                quantile=quantile,
                threshold=threshold,
                parent_weights=parent_weights,
                parent_logic_type='And',
                depth=0,
                explain_type=explain_type,
                print_type=print_type,
                input_features=x[i].cpu(),
                channel=class_prediction,
                force_negate=force_negate,
                ignore_uninformative=ignore_uninformative,
                rounding_precision=rounding_precision,
                inverse_transform=inverse_transform,
                show_bounds=show_bounds
            )

    def predict_proba(self, dl: DataLoader) -> Tuple[pd.DataFrame, pd.DataFrame]:
        predictions, all_targets = self.predict(dl)
        if self.logits:
            predictions.iloc[:, :] = predictions.iloc[:, :] = expit(predictions.values)
        return predictions, all_targets


__all__ = [ReasoningNetworkClassifierMixin]



from copy import deepcopy
from collections import defaultdict
from typing import List, Tuple, Union

import numpy as np
import numpy.typing as npt

import torch
from torch.utils.data import DataLoader
from torchvision.transforms import v2
from pytorch_optimizer import Lookahead

from nrn.models.base import BaseBanditNRNModel
from .base import BaseReasoningNetworkDistributedTrainer


class BanditNRNTrainer(BaseReasoningNetworkDistributedTrainer):

    def __init__(
            self,
            model: BaseBanditNRNModel,
            loss_func,
            optimizer,
            scheduler=None,
            epochs: int = 100,
            accumulation_steps: int = 1,
            objective: str = 'maximize',
            l1_lambda: float = 1e-5,
            lookahead_steps: int = 5,
            lookahead_steps_size: float = 0.7,
            augment: str = None,
            augment_alpha: float = 0.03,
            class_independent: bool = False,
            early_stopping_plateau_count: int = 20,
            perform_prune_plateau_count: int = 10,
            increase_prune_plateau_count: int = 20,
            increase_prune_plateau_count_plateau_count: int = 20
    ):
        """
        Train a BanditRRN model.

        Example:
            trainer = BanditRRNTrainer(
                model=model,
                loss_func=nn.CrossEntropyLoss(),
                optimizer=optim.Adam,
                epochs=100,
                learning_rate=learning_rate,
                accumulation_steps=1,
                l1_lambda=l1_lambda,
                weight_decay=weight_decay,
                early_stopping_plateau_count=early_stopping_plateau_count,
                perform_prune_plateau_count=perform_prune_plateau_count,
                explain_percentile=0.9,
                verbose=0,
                distributed=False
            )

            trainer.train(train_dl, val_dl)

        Args:
            model (BanditRRNModel): the model to be trained.
                Format:
                    This model must implement the following methods:
                        predict(DataLoader): returns Tuple[pd.DataFrame, pd.DataFrame]: prediction values, target values
                        evaluate(DataLoader): returns float: evaluation metric
            loss_func (Loss): One of pytorch's loss classes
            optimizer (Optimizer): One of pytorch's optimizer classes
            scheduler (Scheduler): One of pytorch's scheduler classes
            class_independent (bool): If true, perform pruning and policy updates independently per output.
                Used in most multi-output cases.
            epochs (int): Maximum number of epochs to train for
            l1_lambda (float): L1 regularization weight
            early_stopping_plateau_count (int): Number of epochs without improvement to end training.
            perform_prune_plateau_count (int): Number of epochs without improvement to prune and grow model.
            accumulation_steps (int): Number of optimization steps to perform before backward operation.
            objective (str): 'maximize' or 'minimize'.  Default is 'maximize'.
        """
        if class_independent:
            setattr(loss_func, "reduction", 'none')

        super(BanditNRNTrainer, self).__init__(
            model=model,
            loss_func=loss_func,
            optimizer=optimizer,
            scheduler=scheduler,
            epochs=epochs,
            l1_lambda=l1_lambda,
            early_stopping_plateau_count=early_stopping_plateau_count,
            accumulation_steps=accumulation_steps,
            objective=objective,
        )
        self.prune_and_grow_plateau_count = perform_prune_plateau_count
        self.increase_prune_plateau_count = increase_prune_plateau_count + perform_prune_plateau_count
        self.increase_prune_plateau_count_plateau_count = increase_prune_plateau_count_plateau_count

        self.lookahead_steps = lookahead_steps
        self.lookahead_steps_size = lookahead_steps_size
        if self.lookahead_steps > 0:
            self.optimizer = Lookahead(self.optimizer, k=self.lookahead_steps, alpha=self.lookahead_steps_size)

        self.augment = augment
        self.augment_alpha = augment_alpha
        if self.augment == 'CM':
            self.augmentor = v2.CutMix(num_classes=2, alpha=self.augment_alpha)
        elif self.augment == 'MU':
            self.augmentor = v2.MixUp(num_classes=2, alpha=self.augment_alpha)

        self.class_independent = class_independent
        if self.class_independent:
            self.best_class_val_performances = np.array([np.nan] * self.model.output_size)
            self.best_class_train_performances = np.array([np.nan] * self.model.output_size)
        self.model.best_state['rn'] = deepcopy(self.model.rn)
        self.model.best_state['prune_rn'] = deepcopy(self.model.rn)
        self.model.best_state['epoch'] = 0
        self.model.best_state['prune_epoch'] = 0
        # self.model.best_state['initialized_optimizer'] = deepcopy(self.optimizer)
        # self.model.best_state['prune_initialized_optimizer'] = deepcopy(self.optimizer)
        self.model.best_state['was_pruned'] = False
        self.model.best_state['prune_was_pruned'] = False

    def save_best_state(self, classes: npt.NDArray = None, mode: str = 'prune'):
        """
        Save the state of the network.

        Args:
            classes (List[int]): List of integers indicating which class states to save.
            mode (str): 'eval' or 'prune'.  Determines which set of states to save.  'eval' is for early stopping
                and the final model.  'prune' is for the pruning process.
        """
        if mode == 'prune':
            state_var = 'prune_'
        elif mode == 'eval':
            state_var = ''
        else:
            raise ValueError("`mode` must be 'prune' or 'eval'.")
        if classes is not None:
            new_state_dict = deepcopy(self.model.best_state[state_var+'rn'].state_dict())
            current_state_dict = deepcopy(self.model.rn.state_dict())
            for k, v in new_state_dict.items():
                if k.find('weights') > -1 or k.find('mask') > -1:
                    for c in classes:
                        new_state_dict[k][c].data.copy_(current_state_dict[k][c].data)
            self.model.best_state[state_var+'rn'].load_state_dict(new_state_dict)
        else:
            self.model.best_state[state_var+'rn'] = deepcopy(self.model.rn)

        self.model.best_state[state_var+'epoch'] = self.epoch
        # self.model.best_state[state_var+'initialized_optimizer'] = deepcopy(self.optimizer)
        self.model.best_state[state_var+'was_pruned'] = self.was_pruned

    def set_best_state(self, mode: str = 'eval') -> bool:
        """
        Set the best state from the last saved.

        Args:
            mode (str): 'eval' or 'prune'.  Determines which set of states to restore.  'eval' is for early stopping
                and the final model.  'prune' is for the pruning process.

        Returns:
            bool: was the state loaded successfully
        """
        if mode == 'prune':
            state_var = 'prune_'
        elif mode == 'eval':
            state_var = ''
        else:
            raise ValueError("`mode` must be 'prune' or 'eval'.")

        self.model.rn.load_state_dict(self.model.best_state[state_var+'rn'].state_dict())
        if mode == 'eval':  # only set the epoch to the best epoch if doing so for evaluation
            self.epoch = self.model.best_state[state_var+'epoch']
        # self.optimizer = deepcopy(self.model.best_state[state_var+'initialized_optimizer'])
        self.was_pruned = self.model.best_state[state_var+'was_pruned']

        return self._validate_state_dicts(
            self.model.rn.state_dict(), self.model.best_state[state_var+'rn'].state_dict())
    
    def _check_improvement(self, performance, mode="eval"):
        """
        Check if performances have improved
        
        Args:
            mode (str): 'eval' or 'prune'.  Determines which set of performances to check. 'eval' is for early stopping
                and the final model.  'prune' is for the pruning process.
        """
        if self.objective == 'maximize' and mode == 'eval':
            if mode == 'eval':
                return performance > self.best_val_performance
            elif mode == 'prune':
                return performance > self.best_train_performance
            else:
                raise ValueError("'mode' must be 'eval' or 'prune'.")
        elif self.objective == 'minimize' or mode == 'prune':
            if mode == 'eval':
                return performance < self.best_val_performance
            elif mode == 'prune':
                return performance < self.best_train_performance
            else:
                raise ValueError("'mode' must be 'eval' or 'prune'.")
        else:
            raise AssertionError("`objective` must be one of 'maximize' or 'minimize'")

    def _check_indices_to_update(self, performance: npt.NDArray, mode='eval') -> npt.NDArray:
        """
        Check which class indices should be updated based on if the performance has improved.

        Args:
            performance (npt.NDArray): Array of performance values by class.
            mode (str): 'eval' or 'prune'.  Determines which set of performances to check. 'eval' is for early stopping
                and the final model.  'prune' is for the pruning process.

        Returns:
            npt.NDArray: class indices that have improved.
        """
        if mode == 'eval':
            best_performances = self.best_class_val_performances
        elif mode == 'prune':
            best_performances = self.best_class_train_performances
        else:
            raise ValueError("'mode' must be 'eval' or 'prune'")

        if self.epoch <= 1 and self.objective == 'maximize' and mode == 'eval':
            condition = ((performance > best_performances)
                         | np.isnan(best_performances))
        elif self.epoch > 1 and self.objective == 'maximize' and mode == 'eval':
            condition = (performance > best_performances)
        elif self.epoch <= 1 and (self.objective == 'minimize' or mode == 'prune'):
            condition = ((performance < best_performances)
                         | np.isnan(best_performances))
        elif self.epoch > 1 and (self.objective == 'minimize' or mode == 'prune'):
            condition = (performance < best_performances)
        else:
            raise AssertionError("`objective` must be one of 'maximize' or 'minimize'")

        return np.where(condition)[0]

    def _class_independent_evaluate_step(
            self,
            performance: npt.NDArray,
            mode: str = 'eval'
    ) -> Tuple[float, npt.NDArray]:
        """
        Evaluate which classes have improved.  Report and save results.

        Args:
            performance (npt.NDArray): Performance by class
            mode (str): 'eval' or 'prune'.  Determines which set of performances to check. 'eval' is for early stopping
                and the final model.  'prune' is for the pruning process.

        Returns:
            Tuple[float, npt.NDArray]: mean performance over classes, classes that improved
        """
        indices_to_update = self._check_indices_to_update(performance, mode=mode)
        if len(indices_to_update) > 0:
            self.logger.info(f"CURRENT {mode.upper()} PERFORMANCE FOR CLASSES {indices_to_update} "
                             f"> BEST VAL PERFORMANCE FOR CLASSES {indices_to_update}")
            self.save_best_state(indices_to_update, mode=mode)

        for c in range(self.model.output_size):
            if c in indices_to_update:
                if mode == 'eval':
                    self.best_class_val_performances[c] = performance[c]
                elif mode == 'prune':
                    self.best_class_train_performances[c] = performance[c]
                else:
                    raise ValueError("'mode' must be 'eval' or 'prune'.")

        if mode == 'eval':
            best_performances = self.best_class_val_performances
        elif mode == 'prune':
            best_performances = self.best_class_train_performances
        else:
            raise ValueError("'mode' must be 'eval' or 'prune'.")

        return np.nanmean(best_performances), indices_to_update

    def _increment_prune_and_grow_plateau_counter(
            self,
            indices_to_update: npt.NDArray,
            prune_and_grow_plateau_counter: dict,
            increase_prune_and_grow_plateau_counter: dict
    ) -> Tuple[dict, dict]:
        """
        Increment the prune and grow plateau counters.

        Args:
            indices_to_update (npt.NDArray): Class indices that improved.
            prune_and_grow_plateau_counter (dict): Plateau counts by class to determine when to prune.
            increase_prune_and_grow_plateau_counter (dict): Plateau counts by class to determine when to
                increase the prune_and_grow_plateau_counter

        Returns:
            dict, dict: prune_and_grow_plateau_counter, increase_prune_and_grow_plateau_counter
        """
        for c in range(self.model.output_size):
            if c in indices_to_update:
                prune_and_grow_plateau_counter[c] = 0
                increase_prune_and_grow_plateau_counter[c] = 0
            else:
                prune_and_grow_plateau_counter[c] += 1
                increase_prune_and_grow_plateau_counter[c] += 1

        return prune_and_grow_plateau_counter, increase_prune_and_grow_plateau_counter

    def _prune_evaluate_step(self, **kwargs) -> Tuple[dict, dict, npt.NDArray, bool]:
        """
        Perform evaluation and increment counters for pruning process.

        Args:
            **kwargs:
                prune_and_grow_plateau_counter (dict): Plateau counts by class to determine when to prune.
                increase_prune_and_grow_plateau_counter (dict): Plateau counts by class to determine when to
                    increase the prune_and_grow_plateau_counter
                train_dl (DataLoader): PyTorch DataLoader with the training data.
                evaluation_metric (sklearn.metrics): Scikit-learn performance metric function.
                multi_class (bool): Use multi-class evaluation approach.

        Returns:
            dict, dict, npt.NDArray, bool: prune_and_grow_plateau_counter, increase_prune_and_grow_plateau_counter, 
                classes that improved, was there any improvement
        """
        prune_and_grow_plateau_counter = kwargs.get("prune_and_grow_plateau_counter")
        increase_prune_and_grow_plateau_counter = kwargs.get("increase_prune_and_grow_plateau_counter")
        train_performance = kwargs.get("total_loss")

        # if in class independent mode then each class must have its own val performances
        if self.class_independent:
            train_performance, indices_to_update = self._class_independent_evaluate_step(
                train_performance, mode='prune')
        else:
            indices_to_update = None

        # check if the val performance is better
        improvement_condition = self._check_improvement(train_performance, mode='prune')

        if self.class_independent:
            prune_and_grow_plateau_counter, increase_prune_and_grow_plateau_counter = (
                self._increment_prune_and_grow_plateau_counter(
                indices_to_update,
                prune_and_grow_plateau_counter,
                increase_prune_and_grow_plateau_counter,
            ))

        if improvement_condition:
            if not self.class_independent:
                prune_and_grow_plateau_counter = 0
                increase_prune_and_grow_plateau_counter = 0
            self.best_train_performance = train_performance
        else:
            if not self.class_independent:
                prune_and_grow_plateau_counter += 1
                increase_prune_and_grow_plateau_counter += 1

        return (prune_and_grow_plateau_counter, increase_prune_and_grow_plateau_counter, indices_to_update,
                improvement_condition)

    def _update_policy(self, **kwargs):
        """
        Perform policy update.
        
        Args:
            **kwargs:
                train_dl (DataLoader): PyTorch DataLoader with the training data.
                output_metric (sklearn.metrics): Scikit-learn performance metric function.
        """
        dl = kwargs.get('train_dl')
        indices_to_update = kwargs.get('indices_to_update')
        output_metric = kwargs.get('output_metric')

        if not self.class_independent:
            self.logger.info(f"UPDATING POLICY.")
            self._rn_to_cuda()
            self.model.update_policy(dl, output_metric=output_metric, objective=self.objective)
            self._rn_to_cuda()
        else:
            self.logger.info(f"UPDATING POLICY FOR CLASSES: {indices_to_update}")
            self._rn_to_cuda()
            self.model.update_policy(dl, indices_to_update, output_metric, objective=self.objective)
            self._rn_to_cuda()

    def _update_policy_step(self, **kwargs):
        """
        Check which classes to update and perform policy update.
        
        Args:
            **kwargs (dict): kwargs
            
        Returns:
            dict: **kwargs
        """

        prune_and_grow_plateau_counter, increase_prune_and_grow_plateau_counter, indices_to_update, \
         improvement_condition = self._prune_evaluate_step(**kwargs)

        kwargs.update({'indices_to_update': indices_to_update})

        # update policy if there was an improvement
        if improvement_condition:
            self._update_policy(**kwargs)

        kwargs.update({
                'prune_and_grow_plateau_counter': prune_and_grow_plateau_counter,
                'increase_prune_and_grow_plateau_counter': increase_prune_and_grow_plateau_counter
        })

        return kwargs

    def _prune_step(self, **kwargs) -> Tuple[dict, dict]:
        """
        Perform pruning of the Reasoning Network.
        
        Args:
            **kwargs (dict): kwargs
            
        Returns:
            dict, dict: prune_and_grow_plateau_counter, increase_prune_and_grow_plateau_counter
        """
        self.set_best_state(mode='prune')
        prune_and_grow_plateau_counter = kwargs.get('prune_and_grow_plateau_counter')
        increase_prune_and_grow_plateau_counter = kwargs.get('increase_prune_and_grow_plateau_counter')

        if not self.class_independent:
            if increase_prune_and_grow_plateau_counter >= self.increase_prune_plateau_count_plateau_count:
                self.prune_and_grow_plateau_count = self.increase_prune_plateau_count
            condition = prune_and_grow_plateau_counter >= self.prune_and_grow_plateau_count
            reason = f"REACHED PLATEAU COUNT {self.prune_and_grow_plateau_count}"
        else:
            plateau_counts = [self.increase_prune_plateau_count
                              if v >= self.increase_prune_plateau_count_plateau_count
                              else self.prune_and_grow_plateau_count
                              for k, v in increase_prune_and_grow_plateau_counter.items()]
            plateaued_classes = [k for c, (k, v) in zip(plateau_counts, prune_and_grow_plateau_counter.items())
                                 if v >= c]
            plateaued_classes_counts = [c for c, (k, v) in zip(plateau_counts, prune_and_grow_plateau_counter.items())
                                        if v >= c]
            condition = bool(plateaued_classes)
            reason = f"REACHED PLATEAU COUNT {plateaued_classes_counts} FOR CLASSES {plateaued_classes}."

        if condition:
            if not self.class_independent:
                self.logger.info(f"PERFORMING PRUNE.  {reason}")
                self.set_best_state(mode='prune')
                self._rn_to_cuda()
                self.model.perform_prune(
                    dl=kwargs.get('train_dl'),
                    output_metric=kwargs.get('output_metric'),
                    objective=self.objective
                )
                self._rn_to_cuda()
                prune_and_grow_plateau_counter = 0
            else:
                self.logger.info(f"PERFORMING PRUNE ON CLASSES: {plateaued_classes}.  {reason}")
                self.set_best_state(mode='prune')
                self._rn_to_cuda()
                self.model.perform_prune(
                    dl=kwargs.get('train_dl'),
                    class_indices=plateaued_classes,
                    output_metric=kwargs.get('output_metric'),
                    objective=self.objective
                )
                self._rn_to_cuda()

                for k in plateaued_classes:
                    prune_and_grow_plateau_counter[k] = 0

            self.was_pruned = True

        return prune_and_grow_plateau_counter, increase_prune_and_grow_plateau_counter

    def _increment_plateau_counter(self, val_performance: Union[npt.NDArray, float], **kwargs) -> int:
        """
        Increment the early stopping plateau counter.
        
        Args:
            val_performance (Union[npt.NDArray, float]): Validation performance by class or overall.
            **kwargs:
                plateau_counter (int): Current plateau count.
                
        Returns:
            int: plateau_counter
        """
        plateau_counter = kwargs.get('plateau_counter')

        # if in class independent mode then each class must have its own val performances
        if self.class_independent:
            val_performance, indices_to_update = self._class_independent_evaluate_step(
                val_performance)

        # check if the val performance is better
        improvement_condition = self._check_improvement(val_performance)

        if improvement_condition:
            self.logger.info(f"CURRENT VALIDATION PERFORMANCE {val_performance} "
                             f"IMPROVED VS BEST VAL PERFORMANCE {self.best_val_performance}.")

            plateau_counter = 0
            if not self.class_independent:
                # best states have already been saved by class in class_independent_evaluate_step
                self.save_best_state(mode='eval')
            self.best_val_performance = val_performance
        else:
            plateau_counter += 1

        return plateau_counter

    def _validation_step(self, **kwargs):
        """
        Perform validation step and increment early stopping plateau counter.

        Args:
            **kwargs:
                val_dl (DataLoader): PyTorch DataLoader with validation data.
                evaluation_metric (sklearn.metrics): Scikit-learn performance metric function.
                multi_class (bool): Use multi-class evaluation approach.

        Returns:
            int: plateau_counter
        """
        epoch = kwargs.get('epoch')

        self.logger.info(f"VALIDATION EVALUATION - EPOCH [{epoch}]: \n")

        predictions, targets = self.model.predict(kwargs.get('val_dl'))
        val_performance = self.model.evaluate(
            predictions=predictions,
            labels=targets,
            class_independent=self.class_independent,
            output_metric=kwargs.get('evaluation_metric'),
            multi_class=kwargs.get('multi_class')
        )

        plateau_counter = self._increment_plateau_counter(val_performance=val_performance, **kwargs)

        return plateau_counter

    def _augment_data(self, features: torch.Tensor, target: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Perform data augmentation.

        Args:
            features (torch.Tensor): features from batch.
            target (torch.Tensor): targets from batch.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: augmented features, augmented targets
        """
        if self.augment in ['CM', 'MU']:
            augmented_features = [features]
            augmented_targets = [target]
            if self.model.output_size > 1:
                i = np.random.choice(self.model.output_size)
            else:
                i = 0
            new_features, _ = self.augmentor(features.unsqueeze(1).unsqueeze(1), target[:, i])
            augmented_features += [new_features.squeeze(1).squeeze(1)]
            augmented_targets += [target]
            features = torch.cat(augmented_features, dim=0)
            target = torch.cat(augmented_targets, dim=0)

        return features, target

    def _process_batches(
            self,
            dl: DataLoader,
            optimizer,
            **kwargs,
    ) -> Tuple[float, int]:
        """
        Perform batched epoch.

        Args:
            dl (DataLoader): the training dataloader
                Format:
                    DataLoader must return batches with the keys:
                        'features': [BATCH_SIZE, N_FEATURES]
                        'target': [BATCH_SIZE, N_TARGETS]
            optimizer (Optimizer): pytorch optimizer

        Returns:
            Tuple[float, dict, int]: total loss, kwargs, total steps taken
        """
        total_loss = 0
        total_steps = kwargs.get('total_steps')
        for i, batch in enumerate(dl):

            # [BATCH_SIZE, N_FEATURES]
            features = batch['features']
            # [BATCH_SIZE, N_TARGETS]
            target = batch['target'].squeeze()
            target = target.reshape(features.size(0), -1)

            if self.USE_CUDA:
                features = features.cuda()
                target = target.cuda()
            elif self.USE_MPS:
                features = features.to('mps')
                target = target.to('mps')

            # data augmentation
            if self.augment is not None:
                features, target = self._augment_data(features, target)

            # [BATCH_SIZE, N_TARGETS]
            predictions = self.model.rn(features)
            predictions = predictions.squeeze()
            predictions = predictions.reshape(features.size(0), -1)

            # [BATCH_SIZE, N_TARGETS]
            loss = self.loss_func(predictions, target.float())

            # add l1 loss
            if self.l1_lambda > 0.0:
                l1_norm = sum(p.abs().sum() for p in self.model.rn.parameters())
                loss = loss + self.l1_lambda * l1_norm

            loss /= self.accumulation_steps

            if loss.ndim >= 1:
                loss = loss.sum(axis=0)
                loss_agg = loss.mean()
            else:
                loss_agg = loss

            loss_agg.backward()

            if (i + 1) % self.accumulation_steps == 0 or (i + 1 == len(dl)):
                optimizer.step()
                # if using stochastic weight averaging
                if self.model.averaged_rn is not None:
                    self.model.averaged_rn.update_parameters(self.model.rn)
                optimizer.zero_grad()
                if self.scheduler is not None:
                    self.scheduler.step()
                total_steps += 1

            # report aggregated loss if using distributed training
            total_loss += loss

        if isinstance(total_loss, torch.Tensor):
            total_loss = total_loss.detach().cpu().numpy() / len(dl.dataset)

        return total_loss, total_steps

    def train(self, train_dl: DataLoader, val_dl: DataLoader = None, evaluation_metric=None, multi_class=False):
        """
        Train the rn model.

        Args:
            train_dl (DataLoader): data loader for training data
            val_dl (DataLoader): data loader for a held-out set used for early stopping
            evaluation_metric (function): scikit-learn evaluation metrics
            multi_class (bool): treat problem as multi-class for evaluation purposes

                Format:
                    DataLoaders must return batches with the keys:
                        'features': [BATCH_SIZE, N_FEATURES]
                        'target': [BATCH_SIZE, N_TARGETS]
        """
        total_steps = 0
        total_loss = 1e12
        plateau_counter = 0

        if self.class_independent:
            prune_and_grow_plateau_counter = defaultdict(float)
            increase_prune_and_grow_plateau_counter = defaultdict(float)
        else:
            prune_and_grow_plateau_counter = 0
            increase_prune_and_grow_plateau_counter = 0

        if val_dl is None:
            val_dl = train_dl

        for epoch in range(self.epochs):

            self.epoch = epoch

            kwargs = {
                'val_dl': val_dl,
                'train_dl': train_dl,
                'optimizer': self.optimizer,
                'epoch': epoch,
                'plateau_counter': plateau_counter,
                'prune_and_grow_plateau_counter': prune_and_grow_plateau_counter,
                'increase_prune_and_grow_plateau_counter': increase_prune_and_grow_plateau_counter,
                'evaluation_metric': evaluation_metric,
                'multi_class': multi_class,
                'output_metric': evaluation_metric,
                'total_loss': total_loss
            }

            if self.early_stopping_plateau_count > 0:
                plateau_counter = self._validation_step(**kwargs)

            if epoch > 0:
                kwargs = self._update_policy_step(**kwargs)
                prune_and_grow_plateau_counter, increase_prune_and_grow_plateau_counter = self._prune_step(**kwargs)

            if plateau_counter >= self.early_stopping_plateau_count > 0:
                self.logger.info(f"STOPPING EARLY.  REACHED PLATEAU COUNT OF {plateau_counter}.")
                break

            total_loss, total_steps = (
                self._process_batches(
                    dl=train_dl,
                    optimizer=self.optimizer,
                    val_dl=val_dl,
                    epoch=epoch,
                    plateau_counter=plateau_counter,
                    prune_and_grow_plateau_counter=prune_and_grow_plateau_counter,
                    increase_prune_and_grow_plateau_counter=increase_prune_and_grow_plateau_counter,
                    total_steps=total_steps,
                    output_metric=evaluation_metric
                ))

            self.logger.info(f"Epoch: {epoch}, Total Loss: {total_loss}")
            epoch += 1

        if self.early_stopping_plateau_count == 0:
            self.save_best_state(mode='eval')


__all__ = [BanditNRNTrainer]



from setuptools import setup, find_packages

def parse_requirements(path):
    with open(path) as f:
        lines = f.read().splitlines()
        return [line.strip() for line in lines if line.strip() and not line.startswith("#")]

if __name__ == "__main__":
    setup(
        name="centaur-nrn",
        version="0.1.3",
        description="A PyTorch framework for rapidly developing Neural Reasoning Networks.NRN module under Centaur organization",
        author="Centaur Team",
        author_email="team@centaur.org",
        url="https://github.com/centaur/nrn",
        project_urls={
        "Documentation": "https://docs.centaur.org/nrn",
        "Source": "https://github.com/centaur/nrn",
        "Issues": "https://github.com/centaur/nrn/issues"},
        packages=find_packages(),
        package_dir={"": "."},
        install_requires=parse_requirements("requirements.txt"),
        classifiers=[
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent"
        ],
        include_package_data=True,
    )


from typing import Union
import logging

import torch
from ..utils.operations import val_clamp
from nrn.nn.base import (LukasiewiczChannelBlock, BasePredicates,
                                VariationalLukasiewiczChannelBlock, AttentionLukasiewiczChannelBlock)


class LukasiewiczChannelAndBlock(LukasiewiczChannelBlock):

    def __init__(
            self,
            channels: int,
            in_features: int,
            out_features: int,
            n_selected_features: int,
            parent_weights_dimension: str,
            operands: Union[LukasiewiczChannelBlock, BasePredicates],
            outputs_key: str,
            add_negations: bool = False,
            weight_init: float = 0.2
    ):
        """
        A logical AND channel block.

        Args:
            channels (int): The number of versions of logic.
            in_features (int): The number of features in total from the previous layer or input.
            out_features (int): The number of output features, corresponding to the number of AND nodes for output.
            n_selected_features (int): The number of features to use for input to logic from in_features.
            parent_weights_dimension (str): One of 'out_features', 'channels'.
            operands (Union[LukasiewiczChannelBlock, BasePredicates]): The child logic.
            outputs_key (str): name of outputs used during explanation generation.
            add_negations (bool): add negated logic matching original logics.
        """
        super(LukasiewiczChannelAndBlock, self).__init__(
            channels, in_features, out_features, n_selected_features,
            parent_weights_dimension, operands, 'And', outputs_key, add_negations, weight_init)
        self.logger = logging.getLogger(self.__class__.__name__)

    def forward(self, x: torch.Tensor):
        """
        A channel logical and.

        Args:
            x: intput tensor [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]

        Returns:
            out: output tensor [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        """
        # X: [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]
        # WEIGHTS: [CHANNELS, N_SELECTED_FEATURES, OUT_FEATURES]
        # MASK: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES]
        if x.ndim == 2:
            x = x.unsqueeze(1)
        if x.ndim == 3:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(1).repeat(1, self.channels, self.out_features, 1).unsqueeze(-2)
        if x.ndim == 4:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(-2).repeat(1, 1, self.out_features, 1, 1)

        # X AFTER MASK: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, N_SELECTED_FEATURES]
        x = x.gather(-1, self.mask.unsqueeze(0).expand(x.size(0), *self.mask.shape).unsqueeze(-2))

        # explicit negation
        # WEIGHTS AFTER TRANSPOSE: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES, 1]
        weights = self.weights.transpose(-2, -1).unsqueeze(-1)
        weights_sign = weights.data.detach().clone().greater_equal(torch.tensor(0.0)).float()
        abs_weights = torch.abs(weights)
        pos = (1 - x) @ (abs_weights * weights_sign)
        neg = x @ (abs_weights * (1 - weights_sign))
        tot = pos + neg
        out = val_clamp(self.bias - tot)

        # OUTPUT: [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        out = out.squeeze(-1).transpose(-2, -1)

        return out


class LukasiewiczChannelOrBlock(LukasiewiczChannelBlock):

    def __init__(
            self,
            channels: int,
            in_features: int,
            out_features: int,
            n_selected_features: int,
            parent_weights_dimension: str,
            operands: Union[LukasiewiczChannelBlock, BasePredicates],
            outputs_key: str,
            add_negations: bool = False,
            weight_init: float = 0.2
    ):
        """
        A logical OR channel block.

        Args:
            channels (int): The number of versions of logic.
            in_features (int): The number of features in total from the previous layer or input.
            out_features (int): The number of output features, corresponding to the number of AND nodes for output.
            n_selected_features (int): The number of features to use for input to logic from in_features.
            parent_weights_dimension (str): One of 'out_features', 'channels'.
            operands (Union[LukasiewiczChannelBlock, BasePredicates]): The child logic.
            outputs_key (str): name of outputs used during explanation generation.
            add_negations (bool): add negated logic matching original logics.
        """
        super(LukasiewiczChannelOrBlock, self).__init__(
            channels, in_features, out_features, n_selected_features,
            parent_weights_dimension, operands, 'Or', outputs_key, add_negations, weight_init)

    def forward(self, x: torch.Tensor):
        """
        A channel logical or.

        Args:
            x: intput tensor [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]

        Returns:
            out: output tensor [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        """
        # X: [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]
        # WEIGHTS: [CHANNELS, N_SELECTED_FEATURES, OUT_FEATURES]
        # MASK: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES]
        if x.ndim == 2:
            x = x.unsqueeze(1)
        if x.ndim == 3:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(1).repeat(1, self.channels, self.out_features, 1).unsqueeze(-2)
        if x.ndim == 4:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(-2).repeat(1, 1, self.out_features, 1, 1)

        # X AFTER MASK: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, N_SELECTED_FEATURES]
        x = x.gather(-1, self.mask.unsqueeze(0).expand(x.size(0), *self.mask.shape).unsqueeze(-2))

        # explicit negation
        # WEIGHTS AFTER TRANSPOSE: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES, 1]
        weights = self.weights.transpose(-2, -1).unsqueeze(-1)
        weights_sign = weights.data.detach().clone().greater_equal(torch.tensor(0.0)).float()
        abs_weights = torch.abs(weights)
        pos = x @ (abs_weights * weights_sign)
        neg = (1 - x) @ (abs_weights * (1 - weights_sign))
        tot = pos + neg
        out = val_clamp(1 - self.bias + tot)

        # OUTPUT: [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        out = out.squeeze(-1).transpose(-2, -1)

        return out


class LukasiewiczChannelXOrBlock(LukasiewiczChannelBlock):

    def __init__(
            self,
            channels: int,
            in_features: int,
            out_features: int,
            n_selected_features: int,
            parent_weights_dimension: str,
            operands: Union[LukasiewiczChannelBlock, BasePredicates],
            outputs_key: str,
            add_negations: bool = False,
            weight_init: float = 0.2
    ):
        """
        A logical XOR channel block.

        Args:
            channels (int): The number of versions of logic.
            in_features (int): The number of features in total from the previous layer or input.
            out_features (int): The number of output features, corresponding to the number of AND nodes for output.
            n_selected_features (int): The number of features to use for input to logic from in_features.
            parent_weights_dimension (str): One of 'out_features', 'channels'.
            operands (Union[LukasiewiczChannelBlock, BasePredicates]): The child logic.
            outputs_key (str): name of outputs used during explanation generation.
            add_negations (bool): add negated logic matching original logics.
        """
        super(LukasiewiczChannelXOrBlock, self).__init__(
            channels, in_features, out_features, n_selected_features,
            parent_weights_dimension, operands, 'XOr', outputs_key, add_negations, weight_init)

    def forward(self, x: torch.Tensor):
        """
        A channel logical xor.

        Args:
            x: intput tensor [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]

        Returns:
            out: output tensor [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        """
        # X: [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]
        # WEIGHTS: [CHANNELS, N_SELECTED_FEATURES, OUT_FEATURES]
        # MASK: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES]
        if x.ndim == 2:
            x = x.unsqueeze(1)
        if x.ndim == 3:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(1).repeat(1, self.channels, self.out_features, 1).unsqueeze(-2)
        if x.ndim == 4:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(-2).repeat(1, 1, self.out_features, 1, 1)

        # X AFTER MASK: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, N_SELECTED_FEATURES]
        x = x.gather(-1, self.mask.unsqueeze(0).expand(x.size(0), *self.mask.shape).unsqueeze(-2))

        # explicit negation
        # WEIGHTS AFTER TRANSPOSE: [CHANNELS, OUT_FEATURES, 1, N_SELECTED_FEATURES]
        weights = self.weights.transpose(-2, -1).unsqueeze(-2)
        weights_sign = weights.data.detach().clone().greater_equal(torch.tensor(0.0)).float()
        abs_weights = torch.abs(weights)
        pos = (abs_weights * weights_sign) * (1.0 - x)
        neg = (abs_weights * (1.0 - weights_sign)) * x
        tot = pos + neg

        # ensure one true statement
        out_0 = (1.0 - tot)
        out_1 = val_clamp(out_0.gather(-1, out_0.argmax(dim=-1).unsqueeze(-1)))
        out_2 = val_clamp(out_0.sum(dim=-1).unsqueeze(-1) - 1.0)
        out = out_1 - out_2

        # OUTPUT: [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        out = out.squeeze(-1).transpose(-2, -1)

        return val_clamp(out)


class VariationalLukasiewiczChannelAndBlock(VariationalLukasiewiczChannelBlock):

    def __init__(
            self,
            channels: int,
            in_features: int,
            out_features: int,
            n_selected_features: int,
            parent_weights_dimension: str,
            operands: Union[LukasiewiczChannelBlock, BasePredicates],
            outputs_key: str,
            add_negations: bool = False,
            weight_init: float = 0.2,
            var_emb_dim: int = 50,
            var_n_layers: int = 2
    ):
        """
        A logical AND channel block.

        Args:
            channels (int): The number of versions of logic.
            in_features (int): The number of features in total from the previous layer or input.
            out_features (int): The number of output features, corresponding to the number of AND nodes for output.
            n_selected_features (int): The number of features to use for input to logic from in_features.
            parent_weights_dimension (str): One of 'out_features', 'channels'.
            operands (Union[LukasiewiczChannelBlock, BasePredicates]): The child logic.
            outputs_key (str): name of outputs used during explanation generation.
            add_negations (bool): add negated logic matching original logics.
        """
        super(VariationalLukasiewiczChannelAndBlock, self).__init__(
            channels, in_features, out_features, n_selected_features,
            parent_weights_dimension, operands, 'And', outputs_key, add_negations, weight_init,
            var_emb_dim, var_n_layers)
        self.logger = logging.getLogger(self.__class__.__name__)

    def forward(self, x: torch.Tensor):
        """
        A channel logical and.

        Args:
            x: intput tensor [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]

        Returns:
            out: output tensor [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        """
        # X: [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]
        # WEIGHTS: [CHANNELS, N_SELECTED_FEATURES, OUT_FEATURES]
        # MASK: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES]
        if x.ndim == 2:
            x = x.unsqueeze(1)
        if x.ndim == 3:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(1).repeat(1, self.channels, self.out_features, 1).unsqueeze(-2)
        if x.ndim == 4:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(-2).repeat(1, 1, self.out_features, 1, 1)

        # X AFTER MASK: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, N_SELECTED_FEATURES]
        x = x.gather(-1, self.mask.unsqueeze(0).expand(x.size(0), *self.mask.shape).unsqueeze(-2))

        # variational sparsity
        attn = self.sample_mask()

        # explicit negation
        # WEIGHTS AFTER TRANSPOSE: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES, 1]
        weights = (self.weights * attn).transpose(-2, -1).unsqueeze(-1)
        weights_sign = weights.data.detach().clone().greater_equal(torch.tensor(0.0)).float()
        abs_weights = torch.abs(weights)
        pos = (1 - x) @ (abs_weights * weights_sign)
        neg = x @ (abs_weights * (1 - weights_sign))
        tot = pos + neg
        out = val_clamp(self.bias - tot)

        # OUTPUT: [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        out = out.squeeze(-1).transpose(-2, -1)

        return out


class VariationalLukasiewiczChannelOrBlock(VariationalLukasiewiczChannelBlock):

    def __init__(
            self,
            channels: int,
            in_features: int,
            out_features: int,
            n_selected_features: int,
            parent_weights_dimension: str,
            operands: Union[LukasiewiczChannelBlock, BasePredicates],
            outputs_key: str,
            add_negations: bool = False,
            weight_init: float = 0.2,
            var_emb_dim: int = 50,
            var_n_layers: int = 2
    ):
        """
        A logical OR channel block.

        Args:
            channels (int): The number of versions of logic.
            in_features (int): The number of features in total from the previous layer or input.
            out_features (int): The number of output features, corresponding to the number of AND nodes for output.
            n_selected_features (int): The number of features to use for input to logic from in_features.
            parent_weights_dimension (str): One of 'out_features', 'channels'.
            operands (Union[LukasiewiczChannelBlock, BasePredicates]): The child logic.
            outputs_key (str): name of outputs used during explanation generation.
            add_negations (bool): add negated logic matching original logics.
        """
        super(VariationalLukasiewiczChannelOrBlock, self).__init__(
            channels, in_features, out_features, n_selected_features,
            parent_weights_dimension, operands, 'Or', outputs_key, add_negations, weight_init,
            var_emb_dim, var_n_layers)

    def forward(self, x: torch.Tensor):
        """
        A channel logical or.

        Args:
            x: intput tensor [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]

        Returns:
            out: output tensor [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        """
        # X: [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]
        # WEIGHTS: [CHANNELS, N_SELECTED_FEATURES, OUT_FEATURES]
        # MASK: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES]
        if x.ndim == 2:
            x = x.unsqueeze(1)
        if x.ndim == 3:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(1).repeat(1, self.channels, self.out_features, 1).unsqueeze(-2)
        if x.ndim == 4:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(-2).repeat(1, 1, self.out_features, 1, 1)

        # X AFTER MASK: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, N_SELECTED_FEATURES]
        x = x.gather(-1, self.mask.unsqueeze(0).expand(x.size(0), *self.mask.shape).unsqueeze(-2))

        # variational sparsity
        attn = self.sample_mask()

        # explicit negation
        # WEIGHTS AFTER TRANSPOSE: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES, 1]
        weights = (self.weights * attn).transpose(-2, -1).unsqueeze(-1)
        weights_sign = weights.data.detach().clone().greater_equal(torch.tensor(0.0)).float()
        abs_weights = torch.abs(weights)
        pos = x @ (abs_weights * weights_sign)
        neg = (1 - x) @ (abs_weights * (1 - weights_sign))
        tot = pos + neg
        out = val_clamp(1 - self.bias + tot)

        # OUTPUT: [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        out = out.squeeze(-1).transpose(-2, -1)

        return out


class VariationalLukasiewiczChannelXOrBlock(VariationalLukasiewiczChannelBlock):

    def __init__(
            self,
            channels: int,
            in_features: int,
            out_features: int,
            n_selected_features: int,
            parent_weights_dimension: str,
            operands: Union[LukasiewiczChannelBlock, BasePredicates],
            outputs_key: str,
            add_negations: bool = False,
            weight_init: float = 0.2,
            var_emb_dim: int = 50,
            var_n_layers: int = 2
    ):
        """
        A logical XOR channel block.

        Args:
            channels (int): The number of versions of logic.
            in_features (int): The number of features in total from the previous layer or input.
            out_features (int): The number of output features, corresponding to the number of AND nodes for output.
            n_selected_features (int): The number of features to use for input to logic from in_features.
            parent_weights_dimension (str): One of 'out_features', 'channels'.
            operands (Union[LukasiewiczChannelBlock, BasePredicates]): The child logic.
            outputs_key (str): name of outputs used during explanation generation.
            add_negations (bool): add negated logic matching original logics.
        """
        super(VariationalLukasiewiczChannelXOrBlock, self).__init__(
            channels, in_features, out_features, n_selected_features,
            parent_weights_dimension, operands, 'XOr', outputs_key, add_negations, weight_init,
            var_emb_dim, var_n_layers)

    def forward(self, x: torch.Tensor):
        """
        A channel logical xor.

        Args:
            x: intput tensor [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]

        Returns:
            out: output tensor [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        """
        # variational sparsity
        attn = self.sample_mask()

        # WEIGHTS AFTER TRANSPOSE: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES, 1]
        weights = (self.weights * attn).transpose(-2, -1).unsqueeze(-1)
        weights_sign = weights.data.detach().clone().greater_equal(torch.tensor(0.0)).float()
        abs_weights = torch.abs(weights)

        # Disjunction
        x_or = x.clone()
        
        # X: [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]
        # WEIGHTS: [CHANNELS, N_SELECTED_FEATURES, OUT_FEATURES]
        # MASK: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES]
        if x_or.ndim == 2:
            x_or = x_or.unsqueeze(1)
        if x_or.ndim == 3:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x_or = x_or.unsqueeze(1).repeat(1, self.channels, self.out_features, 1).unsqueeze(-2)
        if x_or.ndim == 4:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x_or = x_or.unsqueeze(-2).repeat(1, 1, self.out_features, 1, 1)

        # X AFTER MASK: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, N_SELECTED_FEATURES]
        x_or = x_or.gather(-1, self.mask.unsqueeze(0).expand(x_or.size(0), *self.mask.shape).unsqueeze(-2))

        # explicit negation
        pos = x_or @ (abs_weights * weights_sign)
        neg = (1 - x_or) @ (abs_weights * (1 - weights_sign))
        tot = pos + neg
        out_or = val_clamp(1 - self.bias + tot)

        # OUTPUT: [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        out_or = out_or.squeeze(-1).transpose(-2, -1)

        # Conjunction
        x_and = x.clone()
        # X: [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]
        # WEIGHTS: [CHANNELS, N_SELECTED_FEATURES, OUT_FEATURES]
        # MASK: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES]
        if x_and.ndim == 2:
            x_and = x_and.unsqueeze(1)
        if x_and.ndim == 3:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x_and = x_and.unsqueeze(1).repeat(1, self.channels, self.out_features, 1).unsqueeze(-2)
        if x_and.ndim == 4:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x_and = x_and.unsqueeze(-2).repeat(1, 1, self.out_features, 1, 1)

        # X AFTER MASK: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, N_SELECTED_FEATURES]
        x_and = x_and.gather(-1, self.mask.unsqueeze(0).expand(x_and.size(0), *self.mask.shape).unsqueeze(-2))

        pos = (1 - x_and) @ (abs_weights * weights_sign)
        neg = x_and @ (abs_weights * (1 - weights_sign))
        tot = pos + neg
        out_and = val_clamp(self.bias - tot)

        # OUTPUT: [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        out_and = out_and.squeeze(-1).transpose(-2, -1)

        return val_clamp(out_or - out_and)


# class SparseLukasiewiczChannelXOrBlock(SparseLukasiewiczChannelBlock):
#
#     def __init__(
#             self,
#             channels: int,
#             in_features: int,
#             out_features: int,
#             n_selected_features: int,
#             parent_weights_dimension: str,
#             operands: Union[LukasiewiczLayer, LukasiewiczChannelBlock, BasePredicates],
#             outputs_key: str,
#             add_negations: bool = False,
#             weight_init: float = 0.2,
#             tau: float = 0.2,
#             attn_emb_dim: int = 50
#     ):
#         """
#         A logical XOR channel block.
#
#         Args:
#             channels (int): The number of versions of logic.
#             in_features (int): The number of features in total from the previous layer or input.
#             out_features (int): The number of output features, corresponding to the number of AND nodes for output.
#             n_selected_features (int): The number of features to use for input to logic from in_features.
#             parent_weights_dimension (str): One of 'out_features', 'channels'.
#             operands (Union[LukasiewiczLayer, LukasiewiczChannelBlock, BasePredicates]): The child logic.
#             outputs_key (str): name of outputs used during explanation generation.
#             add_negations (bool): add negated logic matching original logics.
#         """
#         super(SparseLukasiewiczChannelXOrBlock, self).__init__(
#             channels, in_features, out_features, n_selected_features,
#             parent_weights_dimension, operands, 'XOr', outputs_key, add_negations, weight_init,
#             tau, attn_emb_dim)
#
#     def forward(self, x: torch.Tensor):
#         """
#         A channel logical xor.
#
#         Args:
#             x: intput tensor [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]
#
#         Returns:
#             out: output tensor [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
#         """
#         # X: [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]
#         # WEIGHTS: [CHANNELS, N_SELECTED_FEATURES, OUT_FEATURES]
#         # MASK: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES]
#         if x.ndim == 2:
#             x = x.unsqueeze(1)
#         if x.ndim == 3:
#             # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
#             x = x.unsqueeze(1).repeat(1, self.channels, self.out_features, 1).unsqueeze(-2)
#         if x.ndim == 4:
#             # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
#             x = x.unsqueeze(-2).repeat(1, 1, self.out_features, 1, 1)
#
#         # X AFTER MASK: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, N_SELECTED_FEATURES]
#         x = x.gather(-1, self.mask.unsqueeze(0).expand(x.size(0), *self.mask.shape).unsqueeze(-2))
#
#         # variational sparsity
#         attn = self.sample_attn()
#
#         # explicit negation
#         # WEIGHTS AFTER TRANSPOSE: [CHANNELS, OUT_FEATURES, 1, N_SELECTED_FEATURES]
#         weights = (self.weights * attn).transpose(-2, -1).unsqueeze(-2)
#         weights_sign = weights.data.detach().clone().greater_equal(torch.tensor(0.0)).float()
#         abs_weights = torch.abs(weights)
#         pos = (abs_weights * weights_sign) * (1.0 - x)
#         neg = (abs_weights * (1.0 - weights_sign)) * x
#         tot = pos + neg
#
#         # ensure one true statement
#         out_0 = (1.0 - tot)
#         out_1 = val_clamp(out_0.gather(-1, out_0.argmax(dim=-1).unsqueeze(-1)))
#         out_2 = val_clamp(out_0.sum(dim=-1).unsqueeze(-1) - 1.0)
#         out = out_1 - out_2
#
#         # OUTPUT: [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
#         out = out.squeeze(-1).transpose(-2, -1)
#
#         return val_clamp(out)


class AttentionLukasiewiczChannelAndBlock(AttentionLukasiewiczChannelBlock):

    def __init__(
            self,
            channels: int,
            in_features: int,
            out_features: int,
            n_selected_features: int,
            parent_weights_dimension: str,
            operands: Union[LukasiewiczChannelBlock, BasePredicates],
            outputs_key: str,
            add_negations: bool = False,
            weight_init: float = 0.2,
            attn_emb_dim: int = 32,
            attn_n_layers: int = 2
    ):
        """
        A logical AND channel block.

        Args:
            channels (int): The number of versions of logic.
            in_features (int): The number of features in total from the previous layer or input.
            out_features (int): The number of output features, corresponding to the number of AND nodes for output.
            n_selected_features (int): The number of features to use for input to logic from in_features.
            parent_weights_dimension (str): One of 'out_features', 'channels'.
            operands (Union[LukasiewiczChannelBlock, BasePredicates]): The child logic.
            outputs_key (str): name of outputs used during explanation generation.
            add_negations (bool): add negated logic matching original logics.
        """
        super(AttentionLukasiewiczChannelAndBlock, self).__init__(
            channels, in_features, out_features, n_selected_features,
            parent_weights_dimension, operands, 'And', outputs_key, add_negations, weight_init,
            attn_emb_dim, attn_n_layers)
        self.logger = logging.getLogger(self.__class__.__name__)

    def forward(self, x: torch.Tensor):
        """
        A channel logical and.

        Args:
            x: intput tensor [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]

        Returns:
            out: output tensor [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        """
        # X: [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]
        # WEIGHTS: [CHANNELS, N_SELECTED_FEATURES, OUT_FEATURES]
        # MASK: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES]
        if x.ndim == 2:
            x = x.unsqueeze(1)
        if x.ndim == 3:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(1).repeat(1, self.channels, self.out_features, 1).unsqueeze(-2)
        if x.ndim == 4:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(-2).repeat(1, 1, self.out_features, 1, 1)

        # X AFTER MASK: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, N_SELECTED_FEATURES]
        x = x.gather(-1, self.mask.unsqueeze(0).expand(x.size(0), *self.mask.shape).unsqueeze(-2))

        # variational sparsity
        weights_sign_attn = self.weights.data.detach().clone().repeat(
            x.size(0), 1, 1, 1).transpose(-2, -1).unsqueeze(-2).greater_equal(torch.tensor(0.0)).float()
        # pos_attn = (1 - x) * weights_sign_attn
        # neg_attn = x * (1 - weights_sign_attn)
        # x_attn = pos_attn + neg_attn
        attn = self.sample_mask(torch.cat([x, weights_sign_attn], dim=-1))
        attn = attn.squeeze(-2).transpose(-2, -1)
        weights = self.weights.repeat(x.size(0), 1, 1, 1)

        # explicit negation
        # WEIGHTS AFTER TRANSPOSE: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES, 1]
        weights = (weights * attn).transpose(-2, -1).unsqueeze(-1)
        weights_sign = weights.data.detach().clone().greater_equal(torch.tensor(0.0)).float()
        abs_weights = torch.abs(weights)
        pos = (1 - x) @ (abs_weights * weights_sign)
        neg = x @ (abs_weights * (1 - weights_sign))
        tot = pos + neg
        out = val_clamp(self.bias - tot)

        # OUTPUT: [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        out = out.squeeze(-1).transpose(-2, -1)

        return out


class AttentionLukasiewiczChannelOrBlock(AttentionLukasiewiczChannelBlock):

    def __init__(
            self,
            channels: int,
            in_features: int,
            out_features: int,
            n_selected_features: int,
            parent_weights_dimension: str,
            operands: Union[LukasiewiczChannelBlock, BasePredicates],
            outputs_key: str,
            add_negations: bool = False,
            weight_init: float = 0.2,
            attn_emb_dim: int = 32,
            attn_n_layers: int = 2
    ):
        """
        A logical OR channel block.

        Args:
            channels (int): The number of versions of logic.
            in_features (int): The number of features in total from the previous layer or input.
            out_features (int): The number of output features, corresponding to the number of AND nodes for output.
            n_selected_features (int): The number of features to use for input to logic from in_features.
            parent_weights_dimension (str): One of 'out_features', 'channels'.
            operands (Union[LukasiewiczChannelBlock, BasePredicates]): The child logic.
            outputs_key (str): name of outputs used during explanation generation.
            add_negations (bool): add negated logic matching original logics.
        """
        super(AttentionLukasiewiczChannelOrBlock, self).__init__(
            channels, in_features, out_features, n_selected_features,
            parent_weights_dimension, operands, 'Or', outputs_key, add_negations, weight_init,
            attn_emb_dim, attn_n_layers)

    def forward(self, x: torch.Tensor):
        """
        A channel logical or.

        Args:
            x: intput tensor [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]

        Returns:
            out: output tensor [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        """
        # X: [BATCH_SIZE, 1, IN_FEATURES] OR [BATCH_SIZE, CHANNELS, 1, IN_FEATURES]
        # WEIGHTS: [CHANNELS, N_SELECTED_FEATURES, OUT_FEATURES]
        # MASK: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES]
        if x.ndim == 2:
            x = x.unsqueeze(1)
        if x.ndim == 3:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(1).repeat(1, self.channels, self.out_features, 1).unsqueeze(-2)
        if x.ndim == 4:
            # X AFTER TRANSPOSES: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, IN_FEATURES]
            x = x.unsqueeze(-2).repeat(1, 1, self.out_features, 1, 1)

        # X AFTER MASK: [BATCH_SIZE, CHANNELS, OUT_FEATURES, 1, N_SELECTED_FEATURES]
        x = x.gather(-1, self.mask.unsqueeze(0).expand(x.size(0), *self.mask.shape).unsqueeze(-2))

        # variational sparsity
        weights_sign_attn = self.weights.data.detach().clone().repeat(
            x.size(0), 1, 1, 1).transpose(-2, -1).unsqueeze(-2).greater_equal(torch.tensor(0.0)).float()
        # pos_attn = x * weights_sign_attn
        # neg_attn = (1 - x) * (1 - weights_sign_attn)
        # x_attn = pos_attn + neg_attn
        attn = self.sample_mask(torch.cat([x, weights_sign_attn], dim=-1))
        attn = attn.squeeze(-2).transpose(-2, -1)
        weights = self.weights.repeat(x.size(0), 1, 1, 1)

        # explicit negation
        # WEIGHTS AFTER TRANSPOSE: [CHANNELS, OUT_FEATURES, N_SELECTED_FEATURES, 1]
        weights = (weights * attn).transpose(-2, -1).unsqueeze(-1)
        weights_sign = weights.data.detach().clone().greater_equal(torch.tensor(0.0)).float()
        abs_weights = torch.abs(weights)
        pos = x @ (abs_weights * weights_sign)
        neg = (1 - x) @ (abs_weights * (1 - weights_sign))
        tot = pos + neg
        out = val_clamp(1 - self.bias + tot)

        # OUTPUT: [BATCH_SIZE, OUT_CHANNELS, 1, OUT_FEATURES]
        out = out.squeeze(-1).transpose(-2, -1)

        return out
